/* netndis.h
 */

#ifndef _NETNDIS_H
#define _NETNDIS_H

#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/netdevice.h>
#include <linux/wireless.h>
#include <linux/version.h>

/* To enable debug messages */
#if defined(DLDR_DEBUG) && !defined(NETNDIS_DEBUG)
#define NETNDIS_DEBUG		3
#endif

#if (!defined (WIRELESS_EXT)) || (WIRELESS_EXT < 10)
#error "netndis driver requires Wireless extensions v10 or later."
#endif /* (!defined (WIRELESS_EXT)) || (WIRELESS_EXT < 10) */
#define WIRELESS_SPY		// enable iwspy support

#define NETNDIS_MAX_KEY_SIZE	30
#define NETNDIS_MAX_KEYS	4

struct netndis_key {
	u16 len;	/* always stored as little-endian */
	char data[NETNDIS_MAX_KEY_SIZE];
} __attribute__ ((packed));

/********************************************************************/
/* Data types                                                       */
/********************************************************************/

struct header_struct {
	/* 802.3 */
	u8 dest[ETH_ALEN];
	u8 src[ETH_ALEN];
	u16 len;
	/* 802.2 */
	u8 dsap;
	u8 ssap;
	u8 ctrl;
	/* SNAP */
	u8 oui[3];
	u16 ethertype;
} __attribute__ ((packed));

#define RUP_EVEN(a) (((a) + 1) & (~1))

/* 802.2 LLC/SNAP header used for Ethernet encapsulation over 802.11 */
#define encaps_hdr ((u8[]){0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00})

#define ENCAPS_OVERHEAD		(sizeof(encaps_hdr) + 2)

#ifdef MAX_SKB_FRAGS
//#define NETNDIS_ZEROCOPY
#endif


struct netndis_private {
	/* Synchronisation stuff */
	spinlock_t lock;
	wait_queue_head_t wq;
	unsigned volatile long mylock;
	int hw_unavailable;
	unsigned volatile long work_scheduled;
	OSSCHED config_work;
	OSSCHED reset_work;
	OSSCHED stats_work;
	struct semaphore reset_complete_sem;
	int reset_err;
	BOOLEAN reset_addressing;
	POSTHRD osthrd;
	struct timer_list scan_event_timer;
	struct timer_list stats_update_timer;

	/* driver state */
	int open;
	ndiswraplinkstatus_t last_linkstatus;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
	u32 pci_state[64 / sizeof(u32)];
#endif

	/* Net device stuff */
	struct net_device *ndev;
	struct net_device_stats stats;
	struct iw_statistics wstats;

	/* Hardware control variables */
	POS_DEVNODE pDevNode;
	HANDLE miniportAdapter;

	/* Capabilities of the hardware/firmware */
	ULONG capabilities;
	char capabilities_str[64];
	int has_pm;
	//int has_sensitivity;
	u16 channel_mask;
	ULONG max_multicast;

	/* Configuration paramaters */
	unsigned long cfg;
	u32 iw_mode;
	u16 iw_encode, tx_key;
	struct netndis_key keys[NETNDIS_MAX_KEYS];
	unsigned char used_keys[NETNDIS_MAX_KEYS];
	struct iw_param bitratemode;
 	char nick[IW_ESSID_MAX_SIZE+1];
	char desired_essid[IW_ESSID_MAX_SIZE+1];
	u16 frag_thresh;
	ULONG freq;
	u16 ap_density, rts_thresh;
	u16 pm_on, pm_mcast, pm_period, pm_timeout;
	int wpa;
	ndiswrapauthmode_t wpa_authmode;
	ndiswrapencryption_t wpa_encryption;
	int tkip_countermeasures;
	BOOLEAN privacy_filter;

#ifdef WIRELESS_SPY
	int			spy_number;
	u_char			spy_address[IW_MAX_SPY][ETH_ALEN];
	struct iw_quality	spy_stat[IW_MAX_SPY];
#endif

	unsigned long           scan_timestamp; /* Time started to scan */
	unsigned long           print_timestamp;

	/* Configuration dependent variables */
	int promiscuous, mc_count, allmulti;

#ifdef NETNDIS_ZEROCOPY
#define NETNDIS_NUM_PCKTCTX 128
	struct {
		struct page *page;
		int initialcount;
		void *packetctx;
	} packetctxarray[NETNDIS_NUM_PCKTCTX];
#endif

	struct proc_dir_entry *proc_unit_dir;
	struct proc_dir_entry *proc_hwinst;
	struct proc_dir_entry *proc_hwaddr;
	struct proc_dir_entry *proc_drivername;
	struct proc_dir_entry *proc_licstatus;
	struct proc_dir_entry *proc_capabilities;
};

#ifdef NETNDIS_DEBUG
extern int netndis_debug;
#define NETNDISDEBUG(n, args...) do { if (netndis_debug>(n)) printk(KERN_DEBUG args); } while(0)
#else
#define NETNDISDEBUG(n, args...) do { } while (0)
#endif	/* NETNDIS_DEBUG */

#define TRACE_ENTER(devname) NETNDISDEBUG(2, "%s: -> %s()\n", devname, __FUNCTION__);
#define TRACE_EXIT(devname)  NETNDISDEBUG(2, "%s: <- %s()\n", devname, __FUNCTION__);

extern void netndis_init(void);
extern void netndis_exit(void);
extern struct net_device *alloc_netndisdev(POS_DEVNODE pDevNode);
extern int __netndis_up(struct net_device *dev);
extern int __netndis_down(struct net_device *dev);
extern int netndis_stop(struct net_device *dev);
//extern int netndis_reinit_firmware(struct net_device *dev);

/********************************************************************/
/* Locking and synchronization functions                            */
/********************************************************************/

extern inline int netndis_lock(struct netndis_private *priv, BOOLEAN evenif_hw_unavailable)
{
	if(test_and_set_bit(0,&priv->mylock)) {
		//printk(KERN_WARNING"%s: EBUSY\n", __FUNCTION__);
		if(OsGetIrql() == OSIRQL_PASSIVE) {
			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(&priv->wq, &wait);
			while(test_and_set_bit(0,&priv->mylock)) {
				set_current_state(TASK_INTERRUPTIBLE);
				schedule();
				if(signal_pending(current)) {
					set_current_state(TASK_RUNNING);
					remove_wait_queue(&priv->wq, &wait);
					return -EINTR;
				}
			}
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&priv->wq, &wait);
		} else {
			return -EBUSY;
		}
	}
	if (!evenif_hw_unavailable && priv->hw_unavailable) {
		//printk(KERN_DEBUG "netndis_lock() called with hw_unavailable (dev=%p)\n", priv->ndev);
		clear_bit(0,&priv->mylock);
		wake_up(&priv->wq);
		return -EBUSY;
	}
	return 0;
}

extern inline void netndis_unlock(struct netndis_private *priv)
{
	clear_bit(0,&priv->mylock);
	wake_up(&priv->wq);
}

#endif /* _NETNDIS_H */
