/*
 * Copyright (c) 2003-2004 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */

#define OS_NOSTRINGREDEFS

#include "oscompat.h"
#include "osservices.h"
#include "ndiswrap.h"
#include "netndis.h"

#include <linux/pci.h>
#include <linux/kmod.h>
#include <linux/ctype.h>

static struct pci_device_id dldrpci_tbl[] =
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	{PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_CLASS_NETWORK_ETHERNET << 8, 0xffff00, 0},
	{PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_CLASS_NETWORK_OTHER << 8, 0xffff00, 0},
#else
	{0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0}, /* dummy non-matching entry for buggy depmod */
#endif
	{0,}                    /* terminate list */
};


MODULE_DEVICE_TABLE(pci, dldrpci_tbl);

extern struct semaphore dldr_init_mutex;

static int dldrpci_suspend(struct pci_dev *pdev, u32 state)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev = pDevNode->netDev;
	struct netndis_private *priv = ndev->priv;
	ULONG wrapState;
	int err;
	
	printk(KERN_DEBUG "%s: entering sleep mode (state=%d)\n", ndev->name, state);

	wrapState = NdisWrapEnter();

	err = netndis_lock(priv, FALSE);
	if (err) {
		printk(KERN_ERR "%s: hw_unavailable on %s\n", ndev->name, __FUNCTION__);
		NdisWrapLeave(wrapState);
		return err;
	}

	err = __netndis_down(ndev);
	if (err)
		printk(KERN_WARNING "%s: %s: Error %d downing interface\n", ndev->name, __FUNCTION__, err);
	
	netif_device_detach(ndev);

	priv->hw_unavailable++;

	err = NdisWrapStatusToErrno(NdisWrapSetPowerState(priv->miniportAdapter, NDISWRAP_PWRSTATE_D0 + state));
	if (err) {
		printk(KERN_WARNING "%s: transition to state D%d failed (err=%d)\n", ndev->name, state, err);
	} else {
		printk(KERN_DEBUG "%s: changed to state D%d\n", ndev->name, state);
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	pci_save_state(pdev);
#else
	pci_save_state(pdev, priv->pci_state);
#endif
	pci_set_power_state(pdev, state);

	netndis_unlock(priv);

	NdisWrapLeave(wrapState);

	return 0;
}

static int dldrpci_resume(struct pci_dev *pdev)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev = pDevNode->netDev;
	struct netndis_private *priv = ndev->priv;
	ULONG wrapState;
	int err;

	printk(KERN_DEBUG "%s: waking up\n", ndev->name);

	wrapState = NdisWrapEnter();

	pci_set_power_state(pdev, 0);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	pci_restore_state(pdev);
#else
	pci_restore_state(pdev, priv->pci_state);
#endif

	err = NdisWrapStatusToErrno(NdisWrapSetPowerState(priv->miniportAdapter, NDISWRAP_PWRSTATE_D0));
	if (err) {
		printk(KERN_WARNING "%s: transition to state D0 failed (err=%d)\n", ndev->name, err);
	} else {
		printk(KERN_DEBUG "%s: changed to state D0\n", ndev->name);
	}

	err = netndis_lock(priv, TRUE);
	if (err) {
		printk(KERN_ERR "%s: hw_unavailable on %s\n", ndev->name, __FUNCTION__);
		NdisWrapLeave(wrapState);
		return err;
	}

	netif_device_attach(ndev);

	priv->hw_unavailable--;

	if (priv->open && (! priv->hw_unavailable)) {
		err = __netndis_up(ndev);
		if (err)
			printk(KERN_WARNING "%s: %s: Error %d restarting interface\n", ndev->name, __FUNCTION__, err);
	}
	
	netndis_unlock(priv);

	NdisWrapLeave(wrapState);

	return 0;
}

static void __devexit dldrpci_remove (struct pci_dev *pdev)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev;
	ULONG wrapState;

	OsTracePrintf("%s: pDevNode=%p\n", __FUNCTION__, pDevNode);

	if(!pDevNode)
		return;

	down(&dldr_init_mutex);

	wrapState = NdisWrapEnter();

	ndev = pDevNode->netDev;
	pDevNode->netDev = NULL;

	if(ndev) {
		unregister_netdev(ndev);
	}

	NdisWrapHaltAdapter(pDevNode->miniportAdapter);

	pci_set_drvdata(pdev, NULL);

	if(ndev) {
		free_netdev(ndev);
	}
	NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
	kfree(pDevNode);

	pci_release_regions(pdev);
	pci_disable_device(pdev);

	NdisWrapLeave(wrapState);

	up(&dldr_init_mutex);
}


static int __devinit dldrpci_probe (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	int err;
	POS_DEVNODE pDevNode;
	struct net_device *ndev = NULL;
	char buf[100];
    char *argv[5], *envp[3], *p;
	u_int8_t csz;
	ULONG wrapState;

	OsTracePrintf("%s: pdev=%p\n", __FUNCTION__, pdev);

	pDevNode = kmalloc(sizeof(*pDevNode), GFP_KERNEL);
	if(!pDevNode) {
		return -ENOMEM;
	}

	down(&dldr_init_mutex);

	wrapState = NdisWrapEnter();

	memset(pDevNode, 0, sizeof(*pDevNode));
	pDevNode->hwDev = pdev;
	pDevNode->hwModule = THIS_MODULE;
	pDevNode->hwBusType = OSDEVTYPE_PCI;

	/* count : chars in slot_name to determine if fake domain_nr must be added
	 * to make 2.4 instance names identical to 2.6 */
	for(p = pdev->slot_name, csz = 0; *p; p++)
		if(*p == ':')
			csz++;
	snprintf(pDevNode->hwInstName, sizeof(pDevNode->hwInstName), "PCI-%s%s-%04X:%04X-%04X:%04X", csz == 1 ? "0000:" : "", pdev->slot_name, pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);

	err = NdisWrapCreateAdapter(pDevNode, DLDRVARDIR, &pDevNode->miniportAdapter);
	OsTracePrintf("%s: NdisWrapCreateAdapter returned %d miniportAdapter=%p instname=%s pDevNode=%p\n", __FUNCTION__, err, pDevNode->miniportAdapter, pDevNode->hwInstName, pDevNode);
	if(err) {
		pDevNode->miniportAdapter = NULL;
		goto outerr;
	}

	if ( (err = pci_enable_device(pdev)) ) {
		printk(KERN_ERR "%s: pci_enable_device failed (err=%d)\n", pdev->slot_name, err);
		goto outerr;
	}

	if (pci_request_regions(pdev, (char*)(THIS_MODULE->name))) {
		printk(KERN_ERR "%s: pci_request_regions failed\n", pdev->slot_name); 
		pci_disable_device(pdev);
		err = -ENODEV;
		goto outerr;
	}

	pci_set_power_state(pdev, 0);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
	pci_restore_state(pdev, NULL);
#endif

#if 0
	/* XXX enable memory-write-invalidate */
	if (pci_set_mwi(pdev)) {
		printk(KERN_ERR "%s: unable to set MWI\n", pdev->slot_name);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		err = -EIO;
		goto outerr;
	}
#endif

	/* XXX 32-bit addressing only */
	if (pci_set_dma_mask(pdev, 0xffffffff)) {
		printk(KERN_ERR "%s: 32-bit DMA not available\n", pdev->slot_name);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		err = -EIO;
		goto outerr;
	}

	/*
	 * Cache line size is used to size and align various
	 * structures used to communicate with the hardware.
	 */
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &csz);
	if (csz == 0) {
		/*
		 * Linux 2.4.18 (at least) writes the cache line size
		 * register as a 16-bit wide register which is wrong.
		 * We must have this setup properly for rx buffer
		 * DMA to work so force a reasonable value here if it
		 * comes up zero.
		 */
		csz = L1_CACHE_BYTES / sizeof(u_int32_t);
		printk("%s: cache line size not set; forcing %u\n", pdev->slot_name, csz);
		pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, csz);
	}
	/*
	 * The default setting of latency timer yields poor results,
	 * set it to the value used by other systems.  It may be worth
	 * tweaking this setting more.
	 */
	pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0xa8);

	pci_set_drvdata(pdev, pDevNode);

	ndev = alloc_netndisdev(pDevNode);
	if(!ndev) {
		pci_set_drvdata(pdev, NULL);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		err = -ENOMEM;
		goto outerr;
	}

	SET_MODULE_OWNER(ndev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	SET_NETDEV_DEV(ndev, &pdev->dev);
#endif

	pDevNode->netDev = ndev;

	err = register_netdev(ndev);
	if (err) {
		printk(KERN_ERR "%s: Failed to register net device (%d)\n", ndev->name, err);
		pDevNode->netDev = NULL;
		free_netdev(ndev);
		pci_set_drvdata(pdev, NULL);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		goto outerr;
	}

	buf[0] = '\0';
	NdisWrapGetVendorDesc(pDevNode->miniportAdapter, buf, sizeof(buf));

	printk(KERN_DEBUG"%s:%s%s at %s (MAC address %02X:%02X:%02X:%02X:%02X:%02X) ready\n", ndev->name,
			buf[0] ? " ":"", buf,
			pdev->slot_name, 
			ndev->dev_addr[0], ndev->dev_addr[1], ndev->dev_addr[2],
			ndev->dev_addr[3], ndev->dev_addr[4], ndev->dev_addr[5]);

	NdisWrapLeave(wrapState);
	up(&dldr_init_mutex);

    envp [0] = "HOME=/";
    envp [1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR;
    envp [2] = 0;

    argv[0] = DLDRSBINDIR"/dldrconfig";
    argv[1] = "--auto";
    sprintf(buf, "--netif=%s", ndev->name);
    argv[2] = buf;
    argv[3] = "--modalias";
    argv[4] = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = call_usermodehelper(argv[0], argv, envp);
#else
    err = call_usermodehelper(argv[0], argv, envp, 1);
#endif
	if(err && (err != SIGTERM)) {
        printk(KERN_ERR"%s: call_usermodehelper %s returned %d\n", __FUNCTION__, argv[0], err);
    }

	return 0;

outerr:
	if(pDevNode->miniportAdapter)
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);

	NdisWrapLeave(wrapState);
	up(&dldr_init_mutex);
	kfree(pDevNode);
	return err;
}

struct pci_driver dldrpci_driver = {
	.name           = "driverloader",
	.id_table       = dldrpci_tbl,
	.probe          = dldrpci_probe,
	.remove         = __devexit_p(dldrpci_remove),

	.suspend        = dldrpci_suspend,
	.resume         = dldrpci_resume,
};
