/*
 * Copyright (c) 2003-2004 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */

#include "oscompat.h"
#include "osservices.h"
#include "osusb.h"
#include "ndiswrap.h"

#define __KERNEL_SYSCALLS__
static int errno;

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/unistd.h>
#include <asm/fcntl.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgalloc.h>

#include <linux/delay.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <linux/ctype.h>
#include <linux/kmod.h>
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,7) )
#include <linux/completion.h>
#endif
#include <linux/pci.h>
#include <linux/random.h>

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
#include <linux/kallsyms.h>
#include <asm/tlbflush.h>

static spinlock_t tqueue_lock __attribute__((unused)) = SPIN_LOCK_UNLOCKED;
#endif

#ifdef for_each_process
#define sigmask_lock sighand->siglock
#define recalc_sigpending_curtask(t) recalc_sigpending()
#else
#define sighand sig
#define recalc_sigpending_curtask(t) recalc_sigpending(t)
#endif

/********************************************************************/
/* Critical Sections (Mutexes) */

typedef struct {
#ifdef CONFIG_SMP
    spinlock_t spinlock;
#else
	volatile unsigned long lock;
#endif
    unsigned long flags;
    unsigned long nestcnt;
    int owner_cpu;
} CRIT_T;

__shimcall__
HCRIT OsCriticalSectionCreate(void)
{
	CRIT_T *mutex;

	DLDR_ASSERT(OsGetIrql() == OSIRQL_PASSIVE);
	mutex  = OsAllocate(sizeof(CRIT_T));
	if(!mutex) {
		ASSERT(mutex);
		return NULL;
	}

#ifdef CONFIG_SMP
	spin_lock_init(&mutex->spinlock);
#else
	mutex->lock = 0;
#endif
	mutex->nestcnt = 0;

	return ((HCRIT)mutex);
}

__shimcall__
VOID OsCriticalSectionDestroy(IN HCRIT hMutex)
{
	if ( NULL == hMutex )
	{
		ASSERT(hMutex);
		return;
	}
	OsFree(hMutex);
}

__shimcall__
VOID OsCriticalSectionAcquire(IN HCRIT hMutex)
{
	CRIT_T *mutex = (CRIT_T *) hMutex;
	unsigned long flags;
	int cpu;

	ASSERT(hMutex);

	if(!hMutex) /* temporary due to HCF global critical section issue */
		return;

	//OsDebugPrintf("C(%p): %s -> HIGH\n", hMutex, OsIrqlName(OsGetIrql()));
	//print_symbol("called from %s\n", (unsigned long)__builtin_return_address(0));
	local_irq_save(flags);

	cpu = smp_processor_id();

#ifdef CONFIG_SMP
	if(!spin_trylock(&mutex->spinlock))
#else
	if(test_and_set_bit(0,(&mutex->lock)))
#endif
	{

		if(cpu == mutex->owner_cpu) {
			mutex->nestcnt++;
			local_irq_restore(flags);
			//OsDebugPrintf("N(%p): HIGH -> %s\n", hMutex, OsIrqlName(OsGetIrql()));
			return;
		}
#ifdef CONFIG_SMP
		spin_lock(&mutex->spinlock);
#else
		mutex->lock = 1;
#endif
	}

	mutex->owner_cpu = cpu;
	mutex->flags = flags;

	return;
}

__shimcall__
VOID OsCriticalSectionRelease(IN HCRIT hMutex)
{
	CRIT_T *mutex = (CRIT_T *) hMutex;

	ASSERT(hMutex);

	if(!hMutex) /* temporary due to HCF global critical section issue */
		return;

	if(mutex->nestcnt) {
		mutex->nestcnt--;
		//OsDebugPrintf("n(%p): HIGH -> %s\n", hMutex, OsIrqlName(OsGetIrql()));
		return;
	}

#ifdef CONFIG_SMP
	spin_unlock(&mutex->spinlock);
#else
	mutex->lock = 0;
#endif
	local_irq_restore(mutex->flags);
	//OsDebugPrintf("c(%p): HIGH -> %s\n", hMutex, OsIrqlName(OsGetIrql()));
}

/********************************************************************/

__shimcall__
int
OsAtomicTestAndSetBit(int nr, volatile void * addr)
{
	return test_and_set_bit(nr, addr);
}

__shimcall__
void
OsAtomicClearBit(int nr, volatile void * addr)
{
	clear_bit(nr, addr);
}

__shimcall__
OSIRQL OsGetIrql(void)
{
	if(irqs_disabled())
		return OSIRQL_HIGH;

	if(in_irq())
		return OSIRQL_DEVICE;

	if((in_interrupt())
#ifdef in_atomic
			|| (in_atomic())
#endif
			|| (in_softirq())
	  )
		return OSIRQL_DISPATCH;

	return OSIRQL_PASSIVE;
}

static const char *IrqlNames[] = {
	"PASSIVE",
	"APC",
	"DISPATCH",
	"DEVICE",
	"HIGH"
};

__shimcall__
const char *OsIrqlName(OSIRQL Irql)
{
	if((Irql >= OSIRQL_PASSIVE) && (Irql <= OSIRQL_HIGH))
		return IrqlNames[Irql];
	else
		return "INVALID";
}

__shimcall__
OSIRQL OsRaiseIrql(OSIRQL NewIrql, unsigned long *state)
{
	OSIRQL OldIrql = OsGetIrql();

	if(NewIrql != OldIrql) {
		if(NewIrql < OldIrql) {
			OsErrorPrintf("%s: cannot lower IRQL from %s to %s\n", __FUNCTION__, OsIrqlName(OldIrql), OsIrqlName(NewIrql));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			print_symbol("called from %s\n", (unsigned long)__builtin_return_address(0));
#endif
		} else {
#if 0
			OsTestPrintf("%s: %s -> %s %p\n", __FUNCTION__, OsIrqlName(OldIrql), OsIrqlName(NewIrql), state);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			print_symbol("called from %s\n", (unsigned long)__builtin_return_address(0));
#endif
#endif
			switch(NewIrql) {
				case OSIRQL_DISPATCH:
					local_bh_disable();
					break;
				case OSIRQL_DEVICE:
					local_bh_disable();
					if(state && (((struct pci_dev *)state)->irq)) {
						if(OldIrql < OSIRQL_DEVICE)
							disable_irq(((struct pci_dev *)state)->irq);
						else
							disable_irq_nosync(((struct pci_dev *)state)->irq);
					}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
					local_irq_count(smp_processor_id())++;
					//irq_enter(smp_processor_id(), 0);
#else
					irq_enter();
#endif
					break;
				case OSIRQL_HIGH:
					local_irq_save((*state));
					//OsDebugPrintf("%s: local_irq_save state=0x%lx\n", __FUNCTION__, *state);
					break;
				default:
					OsErrorPrintf("%s: unsupported IRQL %d\n", __FUNCTION__, NewIrql);
					break;
			}
		}
	}

	DLDR_ASSERT(OsGetIrql() == NewIrql);

	return OldIrql;
}

__shimcall__
void OsRestoreIrql(OSIRQL OldIrql, unsigned long *state)
{
	OSIRQL CurrentIrql = OsGetIrql();

	if(OldIrql != CurrentIrql) {
		if(OldIrql > CurrentIrql) {
			OsErrorPrintf("%s: cannot raise IRQL from %s to %s\n", __FUNCTION__, OsIrqlName(CurrentIrql), OsIrqlName(OldIrql));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			print_symbol("called from %s\n", (unsigned long)__builtin_return_address(0));
#endif
		} else {
#if 0
			OsTestPrintf("%s: %s -> %s %p\n", __FUNCTION__, OsIrqlName(CurrentIrql), OsIrqlName(OldIrql), state);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			print_symbol("called from %s\n", (unsigned long)__builtin_return_address(0));
#endif
#endif
			switch(CurrentIrql) {
				case OSIRQL_HIGH:
					//OsDebugPrintf("%s: local_irq_restore state=0x%lx\n", __FUNCTION__, *state);
					local_irq_restore((*state));
					break;
				case OSIRQL_DEVICE:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
					local_irq_count(smp_processor_id())--;
					//irq_exit(smp_processor_id(), 0);
#else
					irq_exit();
#endif
					if(state && (((struct pci_dev *)state)->irq))
						enable_irq(((struct pci_dev *)state)->irq);
					/* FALLTHROUGH */
				case OSIRQL_DISPATCH:
					local_bh_enable();
					break;
				default:
					OsErrorPrintf("%s: unsupported IRQL %d\n", __FUNCTION__, CurrentIrql);
					break;
			}
		}
	}
	DLDR_ASSERT(OsGetIrql() == OldIrql);
}

__shimcall__
void
OsSoftInterruptsDisable(void)
{
	local_bh_disable();
}

__shimcall__
void
OsSoftInterruptsEnable(void)
{
	local_bh_enable();
}

__shimcall__
unsigned long
OsInterruptsDisable(void)
{
	unsigned long flags;

	//OsDebugPrintf("%s(): %s -> HIGH\n", __FUNCTION__, OsIrqlName(OsGetIrql()));

    local_irq_save(flags);

	return flags;
}

__shimcall__
void
OsInterruptsRestore(unsigned long flags)
{
    local_irq_restore(flags);
	//OsDebugPrintf("%s(): HIGH -> %s\n", __FUNCTION__, OsIrqlName(OsGetIrql()));
}

/********************************************************************/

#define TICKS_TO_MSECS(ticks) (ticks*(1000/HZ))

static inline long MSECS_TO_TICKS(UINT32 msecs)
{
    long ticks = (msecs * HZ) / 1000;

    // round up to next tick
    if ((!ticks && msecs) || (msecs % (1000/HZ)) >= ((1000/HZ) / 2))
		ticks++;

    return ticks;
}

static time_t epoch = 0;

__shimcall__
UINT32 OsGetSystemTime(void)
{
    struct timeval		timestamp;

    do_gettimeofday(&timestamp);

    return ((timestamp.tv_sec)*1000 + timestamp.tv_usec/1000);
}

__shimcall__ void
OsGetSystemTimeSecUSec(long *s, long *us)
{
    struct timeval		timestamp;

    do_gettimeofday(&timestamp);

	if(s)
		*s = timestamp.tv_sec;
	if(us)
		*us = timestamp.tv_usec;
}

__shimcall__
UINT32 OsGetSystemUpTime(void)
{
#if 0
    struct timeval		timestamp;

    do_gettimeofday(&timestamp);

    return ((timestamp.tv_sec-epoch)*1000 + timestamp.tv_usec/1000);
#else
	return jiffies * (1000 / HZ);
#endif
}

__shimcall__
ULONG OsGetTickCount(void)
{
	return jiffies;
}

__shimcall__
void OsSleep(UINT32 ms)
{
    /* we must schedule() for sleeps >= 10ms since HCF calls us in loops
	 * waiting for asynchronous events.
	 */
	if(ms >= TICKS_TO_MSECS(1)) {
		UINT32 start, end;
		long timeout = MSECS_TO_TICKS(ms);
		//long ort = timeout;

		if(OsGetIrql() != OSIRQL_PASSIVE) {
			if(ms >= 100)
				OsErrorPrintf("%s(%lu): in %s context\n", __FUNCTION__, ms, OsIrqlName(OsGetIrql()));
			//	    ASSERT((OsGetIrql() == OSIRQL_PASSIVE));
			mdelay(ms);
			return;
		}

		start = OsGetSystemTime();
		do {
			set_current_state(TASK_UNINTERRUPTIBLE);
		} while ((timeout = schedule_timeout(timeout)));
		end = OsGetSystemTime() - start;
		if(end < ms) {
			//OsErrorPrintf("OsSleep: short timeout; wanted %ld got %ldms ort=%ld\n", ms, end, ort);
			OsSleep(end);
		}
		//else if(end > ms) {
		//	OsErrorPrintf("OsSleep: timeout %ldms (%ldms too much) ort=%ld\n", ms, end - ms, ort);
		//}
	} else {
		// short delay
		mdelay(ms);
	}
}

__shimcall__
void OsMicroSleep(UINT32 us)
{
	udelay(us);
}

__shimcall__
int OsGetTimerSize(void)
{
	return sizeof(struct timer_list);
}

__shimcall__
void OsInitTimer(HOSTIMER hTimer, __kernelcall__ void (*func)(void *), void * data)
{
	struct timer_list *timer = (struct timer_list *)hTimer;

	memset(timer,0,sizeof(*timer));

	init_timer(timer);
	timer->function = (__kernelcall__ void (*)(unsigned long))func;
	timer->data = (unsigned long)data;
}

__shimcall__
int OsSetTimer(HOSTIMER pTimer, UINT32 msec)
{
	return mod_timer((struct timer_list *)pTimer, jiffies + MSECS_TO_TICKS(msec));
}

__shimcall__
int OsCancelTimer(HOSTIMER pTimer, BOOLEAN sync)
{
	return sync ? del_timer_sync((struct timer_list *)pTimer) : del_timer((struct timer_list *)pTimer);
}

__shimcall__
BOOLEAN OsPendingTimer(HOSTIMER pTimer)
{
	return timer_pending((struct timer_list *)pTimer);
}

/********************************************************************/

#define NUM_PRINTK_BUFS 5

static char printk_buf[NUM_PRINTK_BUFS][1024];

static int get_printk_index(void)
{
	unsigned long flags;
	int i;
	static spinlock_t printk_lock __attribute__((unused)) = SPIN_LOCK_UNLOCKED;
	static int printk_index;

	spin_lock_irqsave(&printk_lock, flags);
	printk_index++;
	if(printk_index == NUM_PRINTK_BUFS)
		printk_index = 0;
	i = printk_index;
	spin_unlock_irqrestore(&printk_lock, flags);

	return i;
}

__shimcall__
void    OsErrorVPrintf       (LPCSTR szFormat, PVOID args)
{
#if defined(DLDR_TEST) || defined(DLDR_TRACE)
	UINT32 dwTime = OsGetSystemTime();
#endif
	int bi = get_printk_index();

	printk_buf[bi][0]= '\0';
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,10) )
	vsprintf(printk_buf[bi], szFormat, args);
#else
	vsnprintf(printk_buf[bi], sizeof(printk_buf[bi]), szFormat, args);
#endif

	printk_buf[bi][sizeof(printk_buf[bi])-1]= '\0';

#if 0
	{
		static int lasthadnl=1;
		if(lasthadnl)
			printk(KERN_ERR"%s", printk_buf[bi]);
		else
			printk("%s", printk_buf[bi]);

		lasthadnl = printk_buf[bi][strlen(printk_buf[bi]) - 1] == '\n';
	}
#else
#if defined(DLDR_TEST) || defined(DLDR_TRACE)
	if(printk_buf[bi][0] == '$')
		printk("%s", printk_buf[bi]+1);
	else
		printk(KERN_ERR"%07lu.%03lu %s: %s", dwTime/1000, dwTime%1000, DLDRNAME, printk_buf[bi]);
#else
	printk(KERN_ERR"%s: %s", DLDRNAME, printk_buf[bi]);
#endif
#endif
}

__shimcall__
void    OsErrorPrintf       (LPCSTR szFormat, ...)
{
	va_list args;

	va_start(args, szFormat);
	OsErrorVPrintf(szFormat, args);
	va_end(args);
}

#ifdef DLDR_DEBUG
__shimcall__
void    OsDebugVPrintf       (LPCSTR szFormat, PVOID args)
{
	UINT32 dwTime = OsGetSystemTime();
	int bi = get_printk_index();

	printk_buf[bi][0]= '\0';
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,10) )
	vsprintf(printk_buf[bi], szFormat, args);
#else
	vsnprintf(printk_buf[bi], sizeof(printk_buf[bi]), szFormat, args);
#endif

	printk_buf[bi][sizeof(printk_buf[bi])-1]= '\0';

#if 0
	printk(KERN_DEBUG"%07lu.%03lu %s", dwTime/1000, dwTime%1000, printk_buf[bi]);
#else
	printk(KERN_DEBUG"%03lu.%03lu %s", (dwTime/1000)%1000, dwTime%1000, printk_buf[bi]);
#endif
}

//int dldrdebug=0;
__shimcall__
void    OsDebugPrintf       (LPCSTR szFormat, ...)
{
	va_list args;

//if(!dldrdebug) return;
	va_start(args, szFormat);
	OsDebugVPrintf(szFormat, args);
	va_end(args);
}
#endif /* DLDR_DEBUG */

/* Convert byte to two ascii-hex characters */
	static void
ctohex(char *buf, short c)
{
	static char hex[] = "0123456789abcdef";

	*buf++ = hex[(c&0xf0)>>4];
	*buf = hex[c&0x0f];
}

/* Print a buffer up to 16 bytes long in formatted hex with ascii
 * translation, e.g.,
 * 0000: 30 31 32 33 34 35 36 37 38 39 3a 3b 3c 3d 3e 3f  0123456789:;<=>?
 */
static void
fmtline(short addr, const unsigned char *buf, short len)
{
	char line[80];
	char *aptr,*cptr;
	unsigned char c;
	int i;

	for(i=0; i< sizeof(line); i++)
		line[i] = ' ';
	ctohex(line,(short)((addr&0xff00) >> 8));
	ctohex(line+2,(short)(addr&0xff));
	aptr = &line[6];
	cptr = &line[55];
	while(len-- != 0){
		c = *buf++;
		ctohex(aptr,(short)(c));
		aptr += 3;
		c &= 0x7f;
#ifndef isprint
#define isprint(c) (((c) >= 0x20) && ((c) < 0x7f))
#endif
		*cptr++ = isprint((c)) ? c : '.';
	}
	*cptr++ = '\0';
	printk("%s\n",  line);
}

/*#define MAX_HEX_DUMP 512*/

__shimcall__
void
OsDumpData(const void *bp, unsigned int length)
{
	unsigned short address = 0;
#ifdef MAX_HEX_DUMP
	unsigned int remain=0;

	if(length > MAX_HEX_DUMP) {
		remain = length - MAX_HEX_DUMP;
		length = MAX_HEX_DUMP;
	}
#endif

//if(!dldrdebug) return;
	while(length > 0) {
		int n;

		n = (length > 16) ? 16 : length;
		fmtline(address,bp,n);
		length -= n;
		address += n;
		bp += n;
	}

#ifdef MAX_HEX_DUMP
	if(remain) {
		printk(KERN_WARNING"%s: skipping %d remaining bytes\n", __FUNCTION__, remain);
	}
#endif
}


typedef struct {
	wait_queue_head_t wq;
	int flag;
} EVENT, *PEVENT;

__shimcall__
void OsEventInit(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, pEvent);

	if(pEvent) {
		init_waitqueue_head(&pEvent->wq);
		pEvent->flag = 0;
	}
}

__shimcall__
HOSEVENT OsEventCreate(void)
{
	HOSEVENT hEvent = OsAllocate(sizeof(EVENT));
   
	OsEventInit(hEvent);

	return hEvent;
}

__shimcall__
void OsEventDestroy(HOSEVENT hEvent)
{
	OsFree(hEvent);
}

#ifndef wq_write_lock_irqsave
#define wq_write_lock_irqsave spin_lock_irqsave
#endif
#ifndef wq_write_unlock_irqrestore
#define wq_write_unlock_irqrestore spin_unlock_irqrestore
#endif

__shimcall__
void OsEventWait(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;
	DECLARE_WAITQUEUE(wait, current);
	unsigned long flags;

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, hEvent);

	if(OsGetIrql() != OSIRQL_PASSIVE) {
		OsErrorPrintf("%s(): in %s context\n", __FUNCTION__, OsIrqlName(OsGetIrql()));

		return;
	}

	DLDR_ASSERT(OsGetIrql() == OSIRQL_PASSIVE);

	//OsDebugPrintf("bef AWQ\n");
	add_wait_queue(&pEvent->wq, &wait);
	while(1) {
		wq_write_lock_irqsave(&pEvent->wq.lock, flags);
		if(pEvent->flag) {
			pEvent->flag = 0;
			//OsDebugPrintf("ret OK\n");
			wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
			break;
		}
		wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
		rmb();
		//OsDebugPrintf("UNINT");
		set_current_state(TASK_UNINTERRUPTIBLE);
		//OsDebugPrintf("ST");
		schedule();
	}

	current->state = TASK_RUNNING;
	//OsDebugPrintf("TRUN");
	remove_wait_queue(&pEvent->wq, &wait);
	//OsDebugPrintf("aft RWQ");
}

// sets event to signalled state
__shimcall__
long OsEventSet(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;
	unsigned long flags;
	int previous_state;

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, hEvent);
	OsFlushWriteBuffers();
	wq_write_lock_irqsave(&pEvent->wq.lock, flags);
	previous_state = pEvent->flag;
	pEvent->flag = 1;
	wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
	wmb();
	//OsDebugPrintf("BWU\n");
	wake_up(&pEvent->wq);
	//OsDebugPrintf("AWU\n");

	return previous_state;
}

// waits until event is set by OsEventSet() call or timeout (ms) has passed
__shimcall__
OSEVENT_WAIT_RESULT OsEventWaitTime(HOSEVENT hEvent, UINT32 timeout)
{
	PEVENT pEvent = (PEVENT)hEvent;
	DECLARE_WAITQUEUE(wait, current);
	unsigned long flags;
	long tmticks = MSECS_TO_TICKS(timeout);
	OSEVENT_WAIT_RESULT res = OSEVENT_WAIT_TIMEOUT;

	//OsDebugPrintf("%s: %p timeout=%ld\n", __FUNCTION__, hEvent, timeout);

	if(!timeout) {
		wq_write_lock_irqsave(&pEvent->wq.lock, flags);
		if(pEvent->flag) {
			pEvent->flag = 0;
			res = OSEVENT_WAIT_OK;
		}
		wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
		return res;
	}

	if(OsGetIrql() != OSIRQL_PASSIVE) {
		OsErrorPrintf("%s(): in %s context\n", __FUNCTION__, OsIrqlName(OsGetIrql()));

		return OSEVENT_WAIT_ERROR;
	}

	DLDR_ASSERT(OsGetIrql() == OSIRQL_PASSIVE);

	//OsDebugPrintf("bef AWQ\n");
	add_wait_queue(&pEvent->wq, &wait);
	do {
		wq_write_lock_irqsave(&pEvent->wq.lock, flags);
		if(pEvent->flag) {
			pEvent->flag = 0;
			//OsDebugPrintf("ret OK\n");
			wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
			res = OSEVENT_WAIT_OK;
			break;
		}
		wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
		rmb();
		//OsDebugPrintf("UNINT");
		set_current_state(TASK_UNINTERRUPTIBLE);
		//OsDebugPrintf("ST");
	} while ( tmticks && (tmticks = schedule_timeout(tmticks)) );

	current->state = TASK_RUNNING;
	//OsDebugPrintf("TRUN");
	remove_wait_queue(&pEvent->wq, &wait);
	//OsDebugPrintf("aft RWQ");

	if(res == OSEVENT_WAIT_TIMEOUT) {
		OsDebugPrintf("%s: %p returning OSEVENT_WAIT_TIMEOUT\n", __FUNCTION__, hEvent);
	}
	return res;
}

// sets event to non-signalled state
__shimcall__
long OsEventClear(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;
	unsigned long flags;
	int previous_state;

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, hEvent);

	wq_write_lock_irqsave(&pEvent->wq.lock, flags);
	previous_state = pEvent->flag;
	pEvent->flag = 0;
	wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);
	wmb();

	return previous_state;
}

__shimcall__
long OsEventState(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;
	unsigned long flags;
	int state;

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, hEvent);

	wq_write_lock_irqsave(&pEvent->wq.lock, flags);
	state = pEvent->flag;
	wq_write_unlock_irqrestore(&pEvent->wq.lock, flags);

	return state;
}


__shimcall__
void OsModuleUseCountInc(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	MOD_INC_USE_COUNT;
#else
	(void)try_module_get(THIS_MODULE);
#endif
}

__shimcall__
void OsModuleUseCountDec(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	MOD_DEC_USE_COUNT;
#else
	if(likely(module_is_live(THIS_MODULE)))
		module_put(THIS_MODULE);
#endif
}

typedef struct tasklet_struct OSDPC, *POSDPC;

__shimcall__
HOSDPC OsDpcCreate(__kernelcall__ void (*func)(void *), void * data)
{
	POSDPC pDpc = (POSDPC) OsAllocate(sizeof(OSDPC));

	if(pDpc) {
		tasklet_init(pDpc, (__kernelcall__ void (*)(unsigned long))func, (unsigned long)data);
	}

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, pDpc);
	return (HOSDPC)pDpc;
}

__shimcall__
void OsDpcDestroy(HOSDPC hDpc)
{
	POSDPC pDpc = (POSDPC)hDpc;

	tasklet_disable(pDpc);

	//OsDebugPrintf("%s: %p\n", __FUNCTION__, hDpc);
	kfree(pDpc);
}

__shimcall__
void OsDpcSchedule(HOSDPC hDpc)
{
	tasklet_schedule((POSDPC)hDpc);
}

__shimcall__
void OsDpcEnable(HOSDPC hDpc)
{
	tasklet_enable((POSDPC)hDpc);
}

__shimcall__
void OsDpcDisable(HOSDPC hDpc)
{
	tasklet_disable((POSDPC)hDpc);
}

__shimcall__
void OsDpcDisableNoSync(HOSDPC hDpc)
{
	tasklet_disable_nosync((POSDPC)hDpc);
}

/********************************************************************/

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 8))
static inline _syscall3(pid_t,waitpid,pid_t,pid,int *,wait_stat,int,options)
#endif

__shimcall__
int OsRunThreadSync(__kernelcall__ int (*func)(void *), void * data)
{
	pid_t pid;
	sigset_t tmpsig;
	int result, status;
   
	OsDebugPrintf("%s: func=%p(%p)\n", __FUNCTION__, func, data);

	if (!current->fs->root) {
		OsErrorPrintf("%s: Root filesystem is not mounted.\n", __FUNCTION__);
		return -EPERM;
	}

	pid = kernel_thread(func, data, 0);
	if(pid < 0) {
		OsErrorPrintf("%s: Forking of kernel thread failed (%d)\n", __FUNCTION__, pid);
		return pid;
	}

	/* Block signals, everything but SIGKILL/SIGSTOP */
	spin_lock_irq(&current->sigmask_lock);
	tmpsig = current->blocked;
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP));
	recalc_sigpending_curtask(current);
	spin_unlock_irq(&current->sigmask_lock);

	{
		mm_segment_t fs = get_fs ();
		set_fs (get_ds ());
		status = 0; /* SuSE 9.0 workaround */
		result = waitpid(pid, &status, __WCLONE);
		set_fs (fs);
	}

	/* Allow signals again */
	spin_lock_irq(&current->sigmask_lock);
	current->blocked = tmpsig;
	recalc_sigpending_curtask(current);
	spin_unlock_irq(&current->sigmask_lock);

	if (result != pid) {
		OsErrorPrintf("%s: Waiting for pid %d failed (errno=%d).\n", __FUNCTION__, pid, -result);
		return result;
	}

	OsDebugPrintf("%s: returned status %d\n", __FUNCTION__, (((status) & 0xff00) >> 8));
	return (((status) & 0xff00) >> 8);
}

typedef struct {
	char *path;
	char **argv;
	char **envp;
} OsExecArgs;

static int OsExec(void *execargs)
{
	OsExecArgs *ea = execargs;
	int err;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	err = exec_usermodehelper(ea->path, ea->argv, ea->envp);
#else
	mm_segment_t fs = get_fs ();
	set_fs (get_ds ());
	err = execve(ea->path, ea->argv, ea->envp);
	set_fs (fs);
	//do_exit(0);
#endif

	if(err) {
		printk(KERN_ERR"%s: exec %s returned: %d\n", __FUNCTION__, ea->path, err);
	}

	return err;
}

__shimcall__
int
OsForkWait(char *program_path, char *argv[], char *envp[])
{
	int err;
	OsExecArgs *ea;
   
	OsDebugPrintf("%s: %s\n", __FUNCTION__, program_path);

	ea = OsAllocate(sizeof(*ea));
	if(!ea)
		return -ENOMEM;

	ea->path = program_path;
	ea->argv = argv;
	ea->envp = envp;

	err = OsRunThreadSync(OsExec, ea);
	OsFree(ea);

	return err;
}

__shimcall__
BOOLEAN OsConfigOption(const char *optionName)
{
	char fn[256];
	FILE *f;

	snprintf(fn, sizeof(fn), "%s/%s", DLDRNVMDIR, optionName);
	fn[sizeof(fn)-1] = '\0';

	f = OsFOpen(fn, "r", NULL);
	if(f) {
		OsFClose(f);
		return TRUE;
	}
	return FALSE;
}

__shimcall__
HOSCONF OsConfigOpen(POS_DEVNODE pDevNode, OSCONFORIGIN origin)
{
	char fn[256];
	FILE *f;

	switch(origin) {
		case OSCONFORIGIN_STATIC:
			snprintf(fn, sizeof(fn), "%s/%02X:%02X:%02X:%02X:%02X:%02X", DLDRNVMDIR,
					pDevNode->hwAddr[0], pDevNode->hwAddr[1],
					pDevNode->hwAddr[2], pDevNode->hwAddr[3],
					pDevNode->hwAddr[4], pDevNode->hwAddr[5]);
			break;
		case OSCONFORIGIN_VENDOR:
			snprintf(fn, sizeof(fn), "%s/%02X:%02X:%02X", DLDRNVMDIR,
					pDevNode->hwAddr[0], pDevNode->hwAddr[1], pDevNode->hwAddr[2]);
			break;
		case OSCONFORIGIN_CONFIG:
			snprintf(fn, sizeof(fn), "%s/%s", DLDRCNFDIR, pDevNode->hwInstName);
			break;
		default:
			return NULL;
	}

	fn[sizeof(fn)-1] = '\0';

	f = OsFOpen(fn, "r", NULL);
	if(!f && (origin == OSCONFORIGIN_CONFIG)) {
		char *argv[] = { DLDRSBINDIR"/dldrconfig", "--auto", "--newinstance", pDevNode->hwInstName, NULL };
		char *envp[] = { "HOME=/", "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR, NULL };

		OsDebugPrintf("%s: cannot open %s\n", __FUNCTION__, fn);

		if(OsForkWait(argv[0], argv, envp))
			return NULL;

		f = OsFOpen(fn, "r", NULL);
		if(!f) {
			OsDebugPrintf("%s: cannot reopen %s\n", __FUNCTION__, fn);
		}
	}

	return f;
}

__shimcall__
HOSCONF OsConfigOpenWrite(POS_DEVNODE pDevNode)
{
	char fn[256];
	FILE *f;
	int err;

	snprintf(fn, sizeof(fn), "%s/%s", DLDRCNFDIR, pDevNode->hwInstName);
	fn[sizeof(fn)-1] = '\0';

	f = OsFOpen(fn, "w", &err);
	if(!f) {
		OsErrorPrintf("%s: cannot open %s for writing (err=%d)\n", __FUNCTION__, fn, err);

		return NULL;
	}

	return f;
}

__shimcall__
void OsConfigParseEthernetAddr(char *string, unsigned char *macaddr)
{
	unsigned char *pb = macaddr;
	char c, hexval[3], *ph = hexval;

	memset(macaddr, 0, 6);

	while(1) {
		c = *string++;

		if(isxdigit(c)) {
			*ph++ = c;
		}

		if(((ph - hexval) == (sizeof(hexval) - 1)) || !isxdigit(c)) {
			if((ph != hexval)) {
				*ph = '\0';
				*pb++ = simple_strtoul(hexval, NULL, 16);
				ph = hexval;
			}
		}

		if(((c == '\n') || (c == '\r') || (c == '\0')) || ((pb - macaddr) == 6))
			break;
	}
}

__shimcall__
int OsConfigWrite(HOSCONF osConfig, char *keyword, POSCONFVALUE confValue, int confSize)
{
	FILE *f = (FILE *)osConfig;
	unsigned char buf[150], *pb = buf, *d, *s;
	int l, err = 0;

	switch(confValue->Type) {
		case osConfTypeInteger:
			snprintf(buf, sizeof(buf), "%s=%ld\n", keyword, confValue->Data.Integer);
			buf[sizeof(buf)-1] = '\0';
			break;
		case osConfTypeHexInteger:
			snprintf(buf, sizeof(buf), "%s=0x%lx\n", keyword, confValue->Data.Integer);
			buf[sizeof(buf)-1] = '\0';
			break;
		case osConfTypeString:
			snprintf(buf, sizeof(buf), "%s=\"%s\"\n", keyword, (PCHAR)confValue->Data.Buffer);
			buf[sizeof(buf)-1] = '\0';
			break;
		case osConfTypeBinary:
			l = strlen(keyword) + 4 + (confSize * 3);

			if(l > sizeof(buf)) {
				pb = OsAllocate(l);
				if(!pb) {
					return -ENOMEM;
				}
			}

			d = pb;
			s = confValue->Data.Buffer;

			d += snprintf(d, l, "%s={", keyword);

			while(confSize > 1) {
				d += snprintf(d, 3+1, "%02X,", *s++);
				confSize--;
			}
			if(confSize == 1)
				d += snprintf(d, 2+1, "%02X", *s++);

			d += snprintf(d, 2+1, "}\n");
			break;
	}

	if(OsFWrite(pb, strlen(pb), 1, f) != 1) {
		err = -EIO;
	}

	if(pb != buf)
		OsFree(pb);

	return err;
}

__shimcall__
int OsConfigRead(HOSCONF osConfig, char *keyword, int keywordsize,
		POSCONFVALUE confValue)
{
	FILE *f = (FILE *)osConfig;
	unsigned char buf[100];
	unsigned char c, *pb;
	char *pk, hexval[10], *ph;
	enum { var, comment, equal, integer, string, binary } state;

	if(!f)
		return -EINVAL;

reset:
	state = var;
	pk = keyword;
	pb = buf;
	ph = hexval;

	while(1) {
		if(OsFRead(&c, sizeof(c), 1, f) != 1) {
			return -ENOSPC; /* EOF */
		}

		if((pk - keyword) == keywordsize) {
			*--pk = '\0';
		}
		if((pb - buf) == sizeof(buf)) {
			*--pb = '\0';
		}
		if((ph - hexval) == sizeof(hexval)) {
			*--ph = '\0';
		}

		if((c == '\n') || (c == '\r') || (c == '\0')) {
			if(state == integer) {
				int negative;
				int base;

				*pb = '\0';
				pb = buf;
				if(*pb == '-') {
					negative = 1;
					pb++;
				} else
					negative = 0;

				if(!strncmp(pb, "0x", 2)) {
					base = 16;
					pb += 2;
					confValue->Type = osConfTypeHexInteger;
				} else {
					base = 10;
					confValue->Type = osConfTypeInteger;
				}

				confValue->Data.Integer = simple_strtoul(pb, NULL, base);;
				if(negative)
					confValue->Data.Integer =  0 - confValue->Data.Integer;
				return 1;
			}

			if((state == string) || (state == binary)) {
				if(state == string) {
					*pb++ = '\0';
					confValue->Type = osConfTypeString;
				} else {
					confValue->Type = osConfTypeBinary;
					if((ph != hexval)) {
						*ph = '\0';
						*pb++ = simple_strtoul(hexval, NULL, 16);
						ph = hexval;
					}
				}

				confValue->Data.Buffer = OsAllocate(pb - buf);
				if(!confValue->Data.Buffer)
					return -ENOMEM;
				memcpy(confValue->Data.Buffer, buf, pb - buf);
				return pb - buf;
			}

			goto reset;
		}

		switch(state) {
			case var:
				if(c == '#') {
					*pk++ = '\0';
					state = comment;
				} else if(c == '=') {
					*pk++ = '\0';
					state = equal;
				} else if(isalnum(c) || (c == '_')) {
					*pk++ = c;
				}
				break;

			case equal:
				if(c == '#') {
					state = comment;
				} else if(c == '{') {
					state = binary;
					ph = hexval;
				} else if(c == '"') {
					state = string;
				} else if(!(isspace(c))) {
					state = integer;
					pb = buf;
					if(isalnum(c) || (c == '-'))
						*pb++ = c;
				}
				break;

			case integer:
				if(c == '#') {
					*pb++ = '\0';
					state = comment;
				} else if(isalnum(c) || (c == '-')) {
					*pb++ = c;
				} else if(!(isspace(c))) {
					/* syntax error */
				}
				break;

			case string:
				if(c == '"') {
					*pb++ = '\0';
					confValue->Type = osConfTypeString;
					confValue->Data.Buffer = OsAllocate(pb - buf);
					if(!confValue->Data.Buffer)
						return -ENOMEM;
					memcpy(confValue->Data.Buffer, buf, pb - buf);
					return pb - buf;
				} else
					*pb++ = c;
				break;

			case binary:
				if(!isxdigit(c) && (ph != hexval)) {
					*ph = '\0';
					*pb++ = simple_strtoul(hexval, NULL, 16);
					ph = hexval;
				}
				if(c == '}') {
					confValue->Type = osConfTypeBinary;
					confValue->Data.Buffer = OsAllocate(pb - buf);
					if(!confValue->Data.Buffer)
						return -ENOMEM;
					memcpy(confValue->Data.Buffer, buf, pb - buf);
					return pb - buf;
				} else if(isxdigit(c)) {
					*ph++ = c;
				}
				break;

			default:
				break;
		}
	}
}

__shimcall__
void OsConfigClose(HOSCONF osConfig)
{
	OsFClose((FILE *)osConfig);
}

__shimcall__
BOOLEAN // whether we have a new instance that should be reinitialized
OsConfigNewInstance(POS_DEVNODE pDevNode)
{
	char mac[20];
	char *argv[] = {
		DLDRSBINDIR"/dldrconfig",
		"--auto",
		"--newinstance",
		pDevNode->hwInstName,
		"--verifymac",
		mac,
		NULL
	};
	char *envp[] = { "HOME=/", "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR, NULL };

	snprintf(mac, sizeof(mac), "%02X:%02X:%02X:%02X:%02X:%02X",
			pDevNode->hwAddr[0], pDevNode->hwAddr[1],
			pDevNode->hwAddr[2], pDevNode->hwAddr[3],
			pDevNode->hwAddr[4], pDevNode->hwAddr[5]);
	mac[sizeof(mac)-1] = '\0';

	return OsForkWait(argv[0], argv, envp) == 123;
}

__shimcall__
void OsFlushWriteBuffers(void)
{
	flush_write_buffers();
	mb();
}

__shimcall__
void OsFlushTlbAll(void)
{
#if 0 //defined(flush_tlb_all) && !defined(CONFIG_X86_PGE) /* XXX */
	//flush_tlb_all();
	//flush_tlb();
#endif
}

__shimcall__
int OsGetTaskStructSize(void)
{
	return sizeof(struct task_struct);
}

__shimcall__
void *OsGetTask(void)
{
	return current;
}

__shimcall__
long OsGetTaskPid(void *task)
{
	return ((struct task_struct *)task)->pid;
}

__shimcall__
char *OsGetTaskCmd(void *task)
{
	return ((struct task_struct *)task)->comm;
}

__shimcall__
int OsGetProcessorId(void)
{
	return smp_processor_id();
}

OS_KERNELCONFIG OsKernelConfig;

__shimcall__
unsigned long OsGetLoopsPerTick(void)
{
	return loops_per_jiffy;
}

__shimcall__
unsigned long OsGetTicksPerSec(void)
{
	return HZ;
}

__shimcall__
unsigned long OsGetTrapDS(void)
{
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
	return __KERNEL_DS;
#else
	return __USER_DS;
#endif
}

__shimcall__
BOOLEAN OsPageFaultTrap(unsigned long error_code)
{
	unsigned long address;

	__asm__("movl %%cr2,%0":"=r" (address));
	if (address >= TASK_SIZE && !(error_code & 5)) {
		pgd_t *pgd, *pgd_k;
		pmd_t *pmd, *pmd_k;
		pte_t *pte_k;
		int index = pgd_index(address);

		asm("movl %%cr3,%0":"=r" (pgd));
		pgd = index + (pgd_t *)__va(pgd);
		pgd_k = init_mm.pgd + index;

		if (!pgd_present(*pgd_k))
			return FALSE;

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
		set_pgd(pgd, *pgd_k);
#endif

		pmd = pmd_offset(pgd, address);
		pmd_k = pmd_offset(pgd_k, address);
		if (!pmd_present(*pmd_k)) 
			return FALSE;
		set_pmd(pmd, *pmd_k);

#ifndef pte_offset_kernel
#define pte_offset_kernel pte_offset
#endif

		pte_k = pte_offset_kernel(pmd_k, address);
		if (!pte_present(*pte_k))
			return FALSE;

		return TRUE; // trap handled
	}

	return FALSE;
}

__shimcall__
int OsGetNumProcessors(void)
{
#ifdef num_online_cpus
	return num_online_cpus();
#else
	return smp_num_cpus;
#endif
}

__shimcall__
void OsGetRandomBytes(void *buf, int nbytes)
{
	get_random_bytes(buf, nbytes);
}

__shimcall__
void OsSeedRandom(unsigned int seed)
{
	net_srandom(seed);
}

struct _OSTHRD {
	const char *name;
	int pid;
	BOOLEAN highestprio;
	task_queue tq;
	wait_queue_head_t wq;
#ifdef DECLARE_COMPLETION
	struct completion started;
	struct completion exited;
#else
	int running;
#endif
	int stop;

	ULONG wrapState;

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
	unsigned long stack_overflows;
	struct task_struct ts;
#endif
};

static void handle_sigs(void)
{
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
	flush_signals(current); /* must be called without spinlock */
#endif
	spin_lock_irq(&current->sigmask_lock);
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
	flush_signals(current); /* must be called with spinlock held */
#endif
	recalc_sigpending_curtask(current);
	spin_unlock_irq(&current->sigmask_lock);
}

static int dldr_thread(OSTHRD *osthrd)
{
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
	int i;
#endif

	osthrd->wrapState = NdisWrapEnter();

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
	daemonize("k%sd/%s", DLDRABBR, osthrd->name);
#ifdef PF_IOTHREAD
	current->flags |= PF_IOTHREAD;
#endif
#endif

	lock_kernel();

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
	exit_mm(current);
	exit_files(current);

	current->session = 1;
	current->pgrp = 1;

	snprintf(current->comm, sizeof(current->comm), "k%sd/%s", DLDRABBR, osthrd->name);
	current->comm[sizeof(current->comm) - 1] = '\0';
#endif

	if(osthrd->highestprio) {
		/* set highest priority */
#ifdef DEF_NICE
		current->nice = -20;
#else
		set_user_nice(current, -20);
#endif

#ifndef MAX_RT_PRIO
#define MAX_RT_PRIO 100
#endif
		current->policy = SCHED_FIFO;
		current->rt_priority = MAX_RT_PRIO-1;
	} else {
#ifdef DEF_NICE
		current->nice = -10;
#else
		set_user_nice(current, -10);
#endif
	}

	unlock_kernel();

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
	flush_signals(current); /* must be called without spinlock */
	allow_signal(SIGTERM);
	allow_signal(SIGKILL);
	allow_signal(SIGSTOP);
#else
	spin_lock_irq(&current->sigmask_lock);
	flush_signals(current); /* must be called with spinlock held */
	{ // flush_signal_handlers
		struct k_sigaction *ka = &current->sighand->action[0];
		for (i = _NSIG ; i != 0 ; i--) {
			ka->sa.sa_handler = SIG_DFL;
			ka->sa.sa_flags = 0;
			sigemptyset(&ka->sa.sa_mask);
			ka++;
		}
	}
	siginitsetinv(&current->blocked, sigmask(SIGTERM) | sigmask(SIGKILL) | sigmask(SIGSTOP));
	recalc_sigpending_curtask(current);
	spin_unlock_irq(&current->sigmask_lock);
#endif

#ifdef DECLARE_COMPLETION
	complete(&osthrd->started);
#else
	osthrd->running = 1;
#endif

	{
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&osthrd->wq, &wait);

		while (!osthrd->stop) {
			set_current_state(TASK_INTERRUPTIBLE);

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
			if(!OsKernelConfig.Stack.Wrap) {
				lock_kernel();
				memcpy(&osthrd->ts.uid, &current->uid, sizeof(osthrd->ts) - (int)(&((struct task_struct *)0)->uid));
				unlock_kernel();
			}
#endif

			run_task_queue(&osthrd->tq);

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
			if(!OsKernelConfig.Stack.Wrap && (current->sighand != osthrd->ts.sighand)) {
				lock_kernel();
				memcpy(&current->uid, &osthrd->ts.uid, sizeof(osthrd->ts) - (int)(&((struct task_struct *)0)->uid));
				unlock_kernel();
				if(!osthrd->stack_overflows++)
					printk(KERN_WARNING"%s: %d stack overflow\n", __FUNCTION__, current->pid);
			}
#endif

			if(signal_pending(current)) {
				handle_sigs();
			} else
				schedule();
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&osthrd->wq, &wait);
	}

	OsThreadExit(osthrd, 0);
    return 0;
}

__shimcall__
void
OsThreadExit(OSTHRD *osthrd, int code)
{
	handle_sigs();

	NdisWrapLeave(osthrd->wrapState);

    osthrd->pid = 0;
#ifdef DECLARE_COMPLETION
#if defined(complete_and_exit) || ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9) )
    complete_and_exit(&osthrd->exited, 0);
#else
    complete(&osthrd->exited);
    up_and_exit(NULL, 0);
#endif
#else
    osthrd->running = 0;
#endif
}

static void
OsThreadStart(OSTHRD *osthrd, const char *name, BOOLEAN highestprio)
{
	memset(osthrd, 0, sizeof(*osthrd));

	osthrd->name = name;
	osthrd->highestprio = highestprio;

	INIT_LIST_HEAD(&osthrd->tq);
	init_waitqueue_head(&osthrd->wq);

#ifdef DECLARE_COMPLETION
	init_completion(&osthrd->started);
	init_completion(&osthrd->exited);

	osthrd->pid = kernel_thread((int (*)(void *))dldr_thread, osthrd, 0);
	if(osthrd->pid < 0)
		return;

	wait_for_completion(&osthrd->started);
#else
	osthrd->running = 0;
	osthrd->pid = kernel_thread((int (*)(void *))dldr_thread, NULL, 0);
	if(osthrd->pid < 0)
		return;
	{
		int i;

		/* wait 5 seconds maximum for thread to start */
		for (i=5000; !osthrd->running && (i > 0); i -= 50) {
			OsSleep(50);
		}
		if(i <= 0) {
			printk(KERN_ERR"%s: giving up on thread\n", __FUNCTION__);
		}
	}
#endif

	OsDebugPrintf("%s: started pid %d (%s)\n", __FUNCTION__, osthrd->pid, osthrd->name);
}

static void
OsThreadStop(OSTHRD *osthrd)
{
	OsDebugPrintf("%s: stopping pid %d\n", __FUNCTION__, osthrd->pid);

	osthrd->stop = 1;
	if(osthrd->pid > 0) {
		int r = kill_proc(osthrd->pid, SIGKILL, 1);
		if(r) {
			printk(KERN_ERR"%s: kill_proc osthrd->pid=%d r=%d\n", __FUNCTION__, osthrd->pid, r);
		}
#ifdef DECLARE_COMPLETION
		wait_for_completion(&osthrd->exited);
#else
		{
			int i;

			/* wait 5 seconds maximum for thread to terminate */
			for (i=5000; osthrd->running && (i > 0); i -= 50) {
				OsSleep(50);
			}
			if(i <= 0) {
				printk(KERN_ERR"%s: giving up on thread\n", __FUNCTION__);
			}
		}
#endif
	}
}

__shimcall__
POSTHRD
OsThreadCreate(const char *name, BOOLEAN highestprio, int *pid)
{
	OSTHRD *osthrd = OsAllocate(sizeof(OSTHRD));

	if(osthrd) {
		OsThreadStart(osthrd, name, highestprio);
		if(pid)
			*pid = osthrd->pid;
	}

	return osthrd;
}

__shimcall__
void
OsThreadDestroy(POSTHRD osthrd)
{
	OsThreadStop(osthrd);
	OsFree(osthrd);
}


__shimcall__
void OsThreadScheduleInit(HOSSCHED hWorkStorage, __kernelcall__ void (*func)(void *), void * data)
{
	struct tq_struct *tqs = (struct tq_struct *)hWorkStorage;

	INIT_TQUEUE(tqs, func, data);
}

__shimcall__
int OsThreadSchedule(POSTHRD osthrd, HOSSCHED hWorkStorage)
{
	int ret;
	struct tq_struct *tqs = (struct tq_struct *)hWorkStorage;

	if(!osthrd || osthrd->pid <= 0) {
		printk(KERN_DEBUG "%s: no thread?!\n", __FUNCTION__);
		return 0;
	}

	OsModuleUseCountInc();
	ret = queue_task(tqs, &osthrd->tq);
	if(!ret)
		OsModuleUseCountDec();
	else
		wake_up(&osthrd->wq);

	return ret;
}

__shimcall__
void OsThreadScheduleDone(void)
{   
    OsModuleUseCountDec();
}

__shimcall__ void OsTaskWakeUp(void *Task)
{
	wake_up_process(Task);
}

__shimcall__ void OsTaskYield(void)
{
#if 0
	if(in_atomic()) {
		printk (KERN_WARNING "%s: in_atomic preempt_count=%d IRQL=%s\n", __FUNCTION__, preempt_count(), OsIrqlName(OsGetIrql()));
	}
#endif
	set_current_state(TASK_UNINTERRUPTIBLE);
	if(signal_pending(current)) {
		handle_sigs();
	} else
		schedule();
	current->state = TASK_RUNNING;
}

__shimcall__
void OsDumpStack(void)
{
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20) )
	dump_stack();
#endif
}

__shimcall__
void OsCalledFrom(void *addr)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	print_symbol("called from %s\n", (unsigned long)addr);
#endif
}

#ifndef CONFIG_REGPARM
__kernelcall__
#if (__GNUC__ == 3 && __GNUC_MINOR__ > 1) || __GNUC__ > 3
__attribute__((used))
#endif
static void regParmUsed(BOOLEAN *regParm)
{
	register void **esp asm("esp") = esp;

	*regParm = ((void *)&regParm == esp);
}
#endif

__shimcall__
void OsExit(void)
{
}

__shimcall__
int OsInit(BOOLEAN *regParm, BOOLEAN sw)
{
	char buf[100];

#ifdef CONFIG_4G_VM_LAYOUT
	OsKernelConfig.CFG_4G_VM_LAYOUT = TRUE;
#else
	OsKernelConfig.CFG_4G_VM_LAYOUT = FALSE;
#endif
#ifdef CONFIG_X86_SWITCH_PAGETABLES
	OsKernelConfig.CFG_X86_SWITCH_PAGETABLES = TRUE;
#else
	OsKernelConfig.CFG_X86_SWITCH_PAGETABLES = FALSE;
#endif
#ifdef CONFIG_PREEMPT
	OsKernelConfig.CFG_PREEMPT = TRUE;
#else
	OsKernelConfig.CFG_PREEMPT = FALSE;
#endif
#ifdef CONFIG_SMP
	OsKernelConfig.CFG_SMP = TRUE;
#else
	OsKernelConfig.CFG_SMP = FALSE;
#endif
#ifdef CONFIG_REGPARM
	OsKernelConfig.CFG_REGPARM = TRUE;
#else
	OsKernelConfig.CFG_REGPARM = FALSE;
#endif

#ifdef CONFIG_REGPARM
	*regParm = OsKernelConfig.CFG_REGPARM;
#else
	regParmUsed(regParm);

	OsKernelConfig.CFG_REGPARM = *regParm;
#endif

	OsKernelConfig.Stack.Size = THREAD_SIZE;
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
	OsKernelConfig.Stack.BottomOffset = sizeof(struct task_struct) + 8;
	OsKernelConfig.Stack.WrapPFH = FALSE;
#else
	OsKernelConfig.Stack.BottomOffset = sizeof(struct thread_info) + 8;
	OsKernelConfig.Stack.WrapPFH = TRUE;
#endif

	OsKernelConfig.Stack.Wrap = sw && OsConfigOption("workarounds/stack") &&
		(OsKernelConfig.Stack.Size < (12 * 1024)) &&
		!OsKernelConfig.CFG_4G_VM_LAYOUT &&
		!OsKernelConfig.CFG_X86_SWITCH_PAGETABLES &&
		!OsKernelConfig.CFG_PREEMPT &&
		!OsKernelConfig.CFG_SMP;

	snprintf(buf, sizeof(buf), "%s: stack=%d/%d/%d", (char*)(THIS_MODULE->name) ? (char*)(THIS_MODULE->name) : "driverloader", OsKernelConfig.Stack.Size, OsKernelConfig.Stack.BottomOffset, OsKernelConfig.Stack.Wrap);
	if(OsKernelConfig.CFG_REGPARM)
		strcat(buf, " REGPARM");
	if(OsKernelConfig.CFG_4G_VM_LAYOUT)
		strcat(buf, " 4G_VM_LAYOUT");
	if(OsKernelConfig.CFG_X86_SWITCH_PAGETABLES)
		strcat(buf, " X86_SWITCH_PAGETABLES");
	if(OsKernelConfig.CFG_PREEMPT)
		strcat(buf, " PREEMPT");
	if(OsKernelConfig.CFG_SMP)
		strcat(buf, " SMP");
	printk(KERN_DEBUG"%s\n", buf);

	if(sizeof(OSTIMER) < sizeof(struct timer_list)) {
		OsErrorPrintf("OSTIMER too small (%d < %d)\n", sizeof(OSTIMER), sizeof(struct timer_list));
		return -ENOSPC;
	}

	if(sizeof(OSSCHED) <= sizeof(struct tq_struct)) {
		OsErrorPrintf("OSSCHED too small (%d < %d)\n", sizeof(OSSCHED), sizeof(struct tq_struct));
		return -ENOSPC;
	}

	{
		struct timeval timestamp;
		do_gettimeofday(&timestamp);
		epoch = timestamp.tv_sec;
	}

	OsUsbInit();

	return 0;
}
