/*
 * Copyright (c) 2003-2004 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */

#include "oscompat.h"
#include "osservices.h"
#include "osusb.h"
#include "ndiswrap.h"

#if defined(CONFIG_USB) || defined(CONFIG_USB_MODULE)
#include <linux/usb.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/pgtable.h>


#undef dbg
#define dbg(format, arg...) { OsDebugPrintf("osusb: " format "\n" , ## arg); }
#undef warn
#define warn(format, arg...) { OsErrorPrintf("osusb: " format "\n" , ## arg); }
//#define dbg(format, arg...) do {} while (0)

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18) )
#define usb_fill_int_urb	FILL_INT_URB
#define usb_fill_bulk_urb	FILL_BULK_URB
#define usb_fill_control_urb	FILL_CONTROL_URB
#endif

#if !defined(URB_ASYNC_UNLINK) && defined(USB_ASYNC_UNLINK)
#define URB_ASYNC_UNLINK USB_ASYNC_UNLINK
#endif

#define NUM_URBS_PER_PIPE   4
#define MAX_URB_BUFSIZE		(20*1024)

typedef struct urb *PURB;

typedef enum {
	ULE_FREE,
	ULE_ALLOCATED,
	ULE_QUEUED,
	ULE_SUBMITTED,
	ULE_SUBMITERR,
	ULE_COMPLETED,
	ULE_ABORTED,
	ULE_FAILED,
	ULE_LAST
} URB_LIST_ENTRY_STATE;

typedef struct {
	__u8 bRequestType;
	__u8 bRequest;
	__u16 wValue;
	__u16 wIndex;
	__u16 wLength;
} __attribute__ ((packed)) USB_CTRLREQUEST;

typedef struct {
	struct list_head entry;
	struct list_head queued_entry;
	struct tagURB_LIST *ul;
	struct tagUSBOSHAL *pUsbOsHal;
	URB_LIST_ENTRY_STATE state;
	PURB volatile urb;
	USB_CTRLREQUEST	devRequest;
	PUINT8 bufferPtr;
	size_t bufferSize;
	PUINT8 userBufferPtr;
	size_t userBufferSize;
	PVOID userContext;
	UINT32	startTime;
} URB_LIST_ENTRY, *PURB_LIST_ENTRY;

typedef struct tagURB_LIST {
	struct list_head urb_list;
	struct list_head queued_list;
	volatile int count[ULE_LAST];
} URB_LIST, *PURB_LIST;

struct tagUSBOSHAL {
	struct usb_device	*pUsbDevice;
	BOOL disconnected;
	POS_DEVNODE pDevNode;
	struct list_head pipe_list;
	PUSBOSPIPE pControlInPipe;
	PUSBOSPIPE pControlOutPipe;
};

struct tagUSBOSPIPE {
	struct list_head entry;
	PUSBOSHAL pUsbOsHal;
	int pipe_os;
	URB_LIST UrbList;
	
	UCHAR bEndpointAddress;
	UCHAR bmAttributes;
	UINT16 wMaxPacketSize;
	UCHAR bInterval;
};

static const char *uleStateNames[] = {
	"FREE",
	"ALLOCATED",
	"QUEUED",
	"SUBMITTED",
	"SUBMITERR",
	"COMPLETED",
	"ABORTED",
	"FAILED",
	"LAST"
};

static void	SubmitNextUrbListEntry(PURB_LIST_ENTRY ule);

static void
ChangeUrbListEntryState(PURB_LIST_ENTRY ule, URB_LIST_ENTRY_STATE newstate)
{
	URB_LIST_ENTRY_STATE oldstate;

	dbg("%s: ule=%p %s -> %s", __FUNCTION__, ule,
			uleStateNames[ule->state],
		   	uleStateNames[newstate]);
#if 0
	{
		int i;

		for(i = 0; i < ULE_LAST; i++) {
			dbg("%s: ul=%p %s=%d", __FUNCTION__, ule->ul,
					uleStateNames[i], ule->ul->count[i]);
		}
	}
#endif

	oldstate = ule->state;
	if(oldstate == newstate)
		return;

	ule->ul->count[ule->state]--;
	ule->state = newstate;
	ule->ul->count[ule->state]++;


	if(oldstate == ULE_QUEUED) {
		list_del(&ule->queued_entry);
	}

	if(newstate == ULE_SUBMITTED) {
		//XXXOsModuleUseCountInc();
	} else if(oldstate == ULE_SUBMITTED) {
		//XXXOsModuleUseCountDec();
		if((newstate == ULE_COMPLETED) || (newstate == ULE_FAILED))
			SubmitNextUrbListEntry(ule);
	}

	if(newstate >= ULE_COMPLETED) {
		ChangeUrbListEntryState(ule, ULE_FREE);
	}
}

static inline BOOL
ErrorIndicatesDisconnect(int err)
{
	switch(err) {
		case -EILSEQ:
		case -ENODEV:
		case -ETIMEDOUT:
		case -EPROTO:
			return TRUE;
		default:
			return FALSE;
	}
}

static int
SubmitUrbListEntry(PURB_LIST_ENTRY ule)
{
	if(!ule->pUsbOsHal->disconnected && !ule->ul->count[ULE_SUBMITTED])  {
		int err;

		ule->startTime = OsGetSystemTime();
		ChangeUrbListEntryState(ule, ULE_SUBMITTED);
		if ((err=usb_submit_urb(ule->urb
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
			, (OsGetIrql() == OSIRQL_PASSIVE) ? GFP_KERNEL : GFP_ATOMIC
#endif
		))) {
			warn("%s: submit failed: %d, pipe=0x%x", __FUNCTION__, err, ule->urb->pipe);
			if(err == -EMSGSIZE) {
				warn("%s: -EMSGSIZE %d, maxpacket=%d EP=%d out=%d", __FUNCTION__, ule->urb->transfer_buffer_length, usb_maxpacket(ule->urb->dev, ule->urb->pipe, usb_pipeout(ule->urb->pipe)), usb_pipeendpoint(ule->urb->pipe), usb_pipeout(ule->urb->pipe));
			}
			if(ErrorIndicatesDisconnect(err)) {
				ule->pUsbOsHal->disconnected = TRUE;
				ChangeUrbListEntryState(ule, ULE_QUEUED);
				list_add_tail(&ule->queued_entry, &ule->ul->queued_list);
			} else {
				ChangeUrbListEntryState(ule, ULE_SUBMITERR);
				return err;
			}
		}
	} else {
		ChangeUrbListEntryState(ule, ULE_QUEUED);
		list_add_tail(&ule->queued_entry, &ule->ul->queued_list);
		dbg("%s: submitted count=%d, added %p to pending queue (count=%d)",
				__FUNCTION__,
				ule->ul->count[ULE_SUBMITTED],
				ule,
				ule->ul->count[ULE_QUEUED]);
	}

	return 0;
}


static void	SubmitNextUrbListEntry(PURB_LIST_ENTRY ule)
{
	if(!ule->pUsbOsHal->disconnected && !ule->ul->count[ULE_SUBMITTED] && ule->ul->count[ULE_QUEUED]) {
		PURB_LIST_ENTRY next_ule;
		int err;

		ASSERT(!list_empty(&ule->ul->queued_list));

		next_ule = list_entry(ule->ul->queued_list.next, URB_LIST_ENTRY, queued_entry);
		dbg("%s: %p", __FUNCTION__, next_ule);

		if((err = SubmitUrbListEntry(next_ule))) {
			unsigned long irqlstate;
			OSIRQL oldIrql = OsGetIrql();

			if(oldIrql < OSIRQL_DISPATCH)
				OsRaiseIrql(OSIRQL_DISPATCH, &irqlstate);

			if(usb_pipecontrol(next_ule->urb->pipe)) {
				USB_CTRLREQUEST dr;
				if(next_ule->urb && next_ule->urb->setup_packet)
					memcpy(&dr, next_ule->urb->setup_packet, sizeof(dr));
				else
					memset(&dr, '\0', sizeof(dr));

				CallbackOsUsbControlTransferCompleted(next_ule->userContext, err, dr.bRequest, le16_to_cpup(&dr.wValue), le16_to_cpup(&dr.wIndex), NULL, 0);
			} else {
				CallbackOsUsbBulkOrIntTransferCompleted(next_ule->userContext, err, NULL, 0);
			}

			ChangeUrbListEntryState(next_ule, ULE_FAILED);
			if(oldIrql < OSIRQL_DISPATCH)
				OsRestoreIrql(oldIrql, &irqlstate);
		}
	}
}

static PURB_LIST_ENTRY
GetFreeUrbFromList(PURB_LIST ul)
{
	struct list_head *e;
	PURB_LIST_ENTRY ule;

	list_for_each (e, &ul->urb_list) {
		ule = list_entry(e, URB_LIST_ENTRY, entry);
		if(ule->state == ULE_FREE) {
			ChangeUrbListEntryState(ule, ULE_ALLOCATED);

			ule->userBufferPtr = NULL;
			ule->userBufferSize = 0;
			ule->startTime = 0;
			dbg("%s: ul=%p ule=%p", __FUNCTION__, ul, ule);
			return ule;
		}

	}
	//warn("%s: no more (ul=%p)", __FUNCTION__, ul);
	return NULL;
}

static PURB_LIST_ENTRY AllocateUrbListEntry(PURB_LIST ul, PUSBOSHAL pUsbOsHal, size_t bufferSize)
{
	PURB_LIST_ENTRY ule;

	ule = (PURB_LIST_ENTRY) kmalloc(sizeof(URB_LIST_ENTRY) + bufferSize, GFP_KERNEL);
	if(!ule) {
		err("%s: kmalloc failed", __FUNCTION__);
		return NULL;
	}

	memset(ule, 0, sizeof(*ule));
	ule->state = ULE_FREE;
	ule->ul = ul;
	ule->pUsbOsHal = pUsbOsHal;
	ule->urb = usb_alloc_urb(0
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
			, GFP_KERNEL
#endif
		);
	if(!ule->urb) {
		err("%s: usb_alloc_urb failed", __FUNCTION__);
		kfree(ule);
		return NULL;
	}

	if(bufferSize) {
		ule->bufferPtr = (PUINT8)(ule + 1);
		ule->bufferSize = bufferSize;
	}

	dbg("%s: ul=%p ule=%p bufsize=%d", __FUNCTION__, ul, ule, bufferSize);
	return ule;
}

static void FreeUrbListEntry(PURB_LIST_ENTRY ule)
{
	dbg("%s: ule=%p", __FUNCTION__, ule);
	usb_free_urb(ule->urb);
	kfree(ule);
}

static void AbortListUrbs(PURB_LIST ul)
{
	struct list_head *e;
	PURB_LIST_ENTRY ule;

	dbg("%s: ul=%p", __FUNCTION__, ul);
	list_for_each (e, &ul->urb_list) {
		ule = list_entry(e, URB_LIST_ENTRY, entry);
		dbg("%s: ule=%p", __FUNCTION__, ule);
		if(ule->state != ULE_FREE) {
			dbg("%s: ule %p (urb %p) was still in state %s\n", __FUNCTION__,
					ule, ule->urb, uleStateNames[ule->state]);
			if(ule->state == ULE_SUBMITTED) {
				dbg("%s: unlinking ule %p (urb %p)\n", __FUNCTION__, ule, ule->urb);
				ule->urb->transfer_flags &= ~URB_ASYNC_UNLINK;
				usb_unlink_urb(ule->urb);
			} else
				ChangeUrbListEntryState(ule, ULE_ABORTED);
		}
	}
}

static void FreeUrbList(PURB_LIST ul)
{
	struct list_head *e, *n;
	PURB_LIST_ENTRY ule;

	dbg("%s: ul=%p", __FUNCTION__, ul);
	AbortListUrbs(ul);
	list_for_each_safe (e, n, &ul->urb_list) {
		ule = list_entry(e, URB_LIST_ENTRY, entry);
		dbg("%s: ule=%p", __FUNCTION__, ule);
		if(ule->state != ULE_FREE) {
			warn("%s: unlinking ule %p (urb %p); was in state %s\n", __FUNCTION__,
					ule, ule->urb, uleStateNames[ule->state]);
			ule->urb->transfer_flags &= ~URB_ASYNC_UNLINK;
			usb_unlink_urb(ule->urb);
		}
		ul->count[ule->state]--;
		list_del(&ule->entry);
		FreeUrbListEntry(ule);
	}
}

static BOOL InitUrbList(PURB_LIST ul, PUSBOSHAL pUsbOsHal, size_t bufferSize, int nurbs)
{
	INIT_LIST_HEAD(&ul->urb_list);
	INIT_LIST_HEAD(&ul->queued_list);

	dbg("%s: ul=%p bufsize=%d nurbs=%d", __FUNCTION__, ul, bufferSize, nurbs);
	while(nurbs--) {
		PURB_LIST_ENTRY ule = AllocateUrbListEntry(ul, pUsbOsHal, bufferSize);
		dbg("%s: ule=%p", __FUNCTION__, ule);
		if(!ule) {
			FreeUrbList(ul);
			return FALSE;
		}
		list_add(&ule->entry, &ul->urb_list);
		ul->count[ULE_FREE]++;
	}

	return TRUE;
}


__shimcall__
INT OsUsbGetDescriptor(PUSBOSHAL pUsbOsHal, UCHAR DescType, UCHAR DescIndex, PVOID TransferBuffer, ULONG TransferBufferLength)
{
	OsDebugPrintf("%s: pUsbOsHal=%p DescType=0x%x DescIndex=0x%x\n", __FUNCTION__, pUsbOsHal, DescType, DescIndex);
	
	return usb_get_descriptor(pUsbOsHal->pUsbDevice, DescType, DescIndex, TransferBuffer, TransferBufferLength);
}


#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
extern struct usb_driver dldrusb_driver;
#endif

__shimcall__
VOID OsUsbClaimInterface(PUSBOSHAL pUsbOsHal, UCHAR InterfaceNumber, UCHAR AlternateSetting)
{
	OsTestPrintf("%s: pUsbOsHal=%p InterfaceNumber=%u AlternateSetting=%u\n", __FUNCTION__, pUsbOsHal, InterfaceNumber, AlternateSetting);

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
	{
		int err;
		struct usb_interface *intf = usb_ifnum_to_if(pUsbOsHal->pUsbDevice, InterfaceNumber);

		if(usb_get_intfdata(intf) != pUsbOsHal->pDevNode) {
			OsTestPrintf("%s: claiming additional interface %d\n", __FUNCTION__, InterfaceNumber);
			err = usb_driver_claim_interface(
					to_usb_driver(pUsbOsHal->pUsbDevice->dev.driver),
					intf,
					pUsbOsHal->pDevNode);
			if(err)
				OsErrorPrintf("%s: usb_driver_claim_interface returned %d\n", __FUNCTION__, err);
		}

#if 0
		err = usb_set_interface(pUsbOsHal->pUsbDevice, InterfaceNumber, AlternateSetting);
		if(err)
			OsErrorPrintf("%s: usb_set_interface returned %d\n", __FUNCTION__, err);
#endif
	}
#else
	{
		struct usb_interface *intf = usb_ifnum_to_if(pUsbOsHal->pUsbDevice, InterfaceNumber);

		if(intf->private_data != pUsbOsHal->pDevNode) {
			OsTestPrintf("%s: claiming additional interface %d\n", __FUNCTION__, InterfaceNumber);
			usb_driver_claim_interface(
					&dldrusb_driver,
					intf,
					pUsbOsHal->pDevNode);
		}
	}
#endif
}

__shimcall__
INT OsUsbSetConfiguration(PUSBOSHAL pUsbOsHal, INT Configuration)
{
	OsTestPrintf("%s: pUsbOsHal=%p Configuration=%d\n", __FUNCTION__, pUsbOsHal, Configuration);
	
#ifndef SET_TIMEOUT
#define SET_TIMEOUT 5
#endif
	return usb_control_msg(pUsbOsHal->pUsbDevice, usb_sndctrlpipe(pUsbOsHal->pUsbDevice, 0),
							  USB_REQ_SET_CONFIGURATION, 0, Configuration, 0, NULL, 0, HZ * SET_TIMEOUT);
}

__shimcall__
INT OsUsbResetPort(PUSBOSHAL pUsbOsHal)
{
#if 0
	struct list_head *h;
	PUSBOSPIPE pUsbOsPipe;
	int err = 0;
	
	OsDebugPrintf("%s\n", __FUNCTION__);
	
	list_for_each (h, &pUsbOsHal->pipe_list) {
		pUsbOsPipe = list_entry(h, USBOSPIPE, entry);

		if(pUsbOsPipe) {
			err = usb_clear_halt(pUsbOsPipe->pUsbOsHal->pUsbDevice, pUsbOsPipe->pipe_os);
			if(err < 0) {
				OsErrorPrintf("%s: pUsbOsPipe=%p/0x%x error %d\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, err);
				return err;
			}
		}
	}
#else
	int err;

	OsTestPrintf("%s\n", __FUNCTION__);

#if defined(ref_to_usb_interface_cache) || ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,7) )
	err = __usb_reset_device(pUsbOsHal->pUsbDevice);
#else
	err = usb_reset_device(pUsbOsHal->pUsbDevice);
#endif
	if (err)
		OsErrorPrintf("%s: Failed to reset device %s (%d)\n", __FUNCTION__, pUsbOsHal->pDevNode->hwInstName, err);

#if 0
	err = usb_reset_configuration(pUsbOsHal->pUsbDevice);
	if (err)
		OsErrorPrintf("%s: Failed to reset configuration %s (%d)\n", __FUNCTION__, pUsbOsHal->pDevNode->hwInstName, err);
#endif

#endif

	return err;
}

__shimcall__
INT OsUsbClearHalt(PUSBOSPIPE pUsbOsPipe)
{
	int err;
	
	OsDebugPrintf("%s: pUsbOsPipe=%p/0x%x\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe ? pUsbOsPipe->pipe_os : 0);
	
	if(!pUsbOsPipe)
		return -EINVAL;
	
	err = usb_clear_halt(pUsbOsPipe->pUsbOsHal->pUsbDevice, pUsbOsPipe->pipe_os);
	if(err < 0) {
		OsErrorPrintf("%s: pUsbOsPipe=%p/0x%x error %d\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, err);
	}
	
	return err;
}

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
static int usbintworkaround;
#endif

static void UsbDataCompletionRoutine(PURB urb
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
											 , struct pt_regs *regs
#endif
											 )
{
	PURB_LIST_ENTRY ule = urb->context;
	int UsbStatus = urb->status;
	PUINT8 pDataReceived = urb->transfer_buffer;
	int length = urb->actual_length;
	URB_LIST_ENTRY_STATE newstate;
	
	OsDebugPrintf("%s: userCtx=%p ule=%p urb=%p UsbStatus=%d length=%d %s\n", __FUNCTION__, ule ? ule->userContext : NULL, ule, urb, UsbStatus, length, usb_pipeint(urb->pipe) ? "PIPEINT" : "");

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
		if (usb_pipeint(urb->pipe)) { // prevent auto-resubmit
			// unfortunately there is no consistent way across the various
			// host controller drivers and their multiple versions to achieve this
			if (usbintworkaround || (ule->pUsbOsHal->pUsbDevice->speed == USB_SPEED_HIGH))
				urb->status = -ENOENT;
			else
				urb->interval = 0;
		}
#endif

	if(ErrorIndicatesDisconnect(UsbStatus)) {
		err("%s: error: URB status %d, surprise disconnect?", __FUNCTION__, UsbStatus);
		ule->pUsbOsHal->disconnected = TRUE;
		ChangeUrbListEntryState(ule, ULE_QUEUED);
		list_add_tail(&ule->queued_entry, &ule->ul->queued_list);
		return;
	}

	switch(UsbStatus) {
		case 0:
			if (urb->pipe & USB_DIR_IN) {

				if(ule->userBufferPtr) {
					memcpy(ule->userBufferPtr, pDataReceived, length);
					pDataReceived = ule->userBufferPtr;
				}
#if 0
				if(usb_pipeint(urb->pipe)) {
					OsDebugDump(pDataReceived, length);
				}
#endif
			}
			newstate = ULE_COMPLETED;
			break;
		case -ECONNRESET:
			dbg("%s: -ECONNRESET", __FUNCTION__);
			newstate = ULE_ABORTED;
			break;
		case -ENOENT:
			dbg("%s: -ENOENT", __FUNCTION__);
			newstate = ULE_ABORTED;
			break;
		case -ESHUTDOWN:
			newstate = ULE_FAILED;
			break;
		default:
			err("%s: error: URB status %d", __FUNCTION__, UsbStatus);
			newstate = ULE_FAILED;
			break;
	}

	CallbackOsUsbBulkOrIntTransferCompleted(ule->userContext, UsbStatus, pDataReceived, length);
	
	ChangeUrbListEntryState(ule, newstate);
}

__shimcall__
INT OsUsbMakeBulkOrIntTransfer(PUSBOSPIPE pUsbOsPipe, PVOID userContext, PVOID TransferBuffer, ULONG TransferBufferLength)
{
	PURB_LIST_ENTRY ule;
	int err;
	
	OsDebugPrintf("%s: pUsbOsPipe=%p/0x%x dir=%s\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, (pUsbOsPipe->pipe_os & USB_DIR_IN) ? "IN" : "OUT");

	ule = GetFreeUrbFromList (&pUsbOsPipe->UrbList);
	if(!ule) {
		OsErrorPrintf("%s: pUsbOsPipe=%p/0x%x no free URBs\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os);

		return -ENOMEM;
	}

	ule->userContext = userContext;

	if (pUsbOsPipe->pipe_os & USB_DIR_IN) {
		if(!IsKMallocMem(TransferBuffer)) {
			ASSERT(nBytes <= ule->bufferSize);
			
			ule->userBufferPtr = TransferBuffer;
			TransferBuffer = ule->bufferPtr;
		} else
			ule->userBufferPtr = NULL;
	} else {
		if(!IsKMallocMem(TransferBuffer)) {
			ASSERT(nBytes <= ule->bufferSize);
			memcpy(ule->bufferPtr, TransferBuffer, TransferBufferLength);
			TransferBuffer = ule->bufferPtr;
		}
	}

	if(usb_pipeint(pUsbOsPipe->pipe_os)) {
		
		usb_fill_int_urb(ule->urb, pUsbOsPipe->pUsbOsHal->pUsbDevice, pUsbOsPipe->pipe_os,
						 TransferBuffer,
						 TransferBufferLength,
						 UsbDataCompletionRoutine,
						 ule,
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
						 pUsbOsPipe->bInterval
#else
						 (pUsbOsPipe->pUsbOsHal->pUsbDevice->speed == USB_SPEED_HIGH) ? pUsbOsPipe->bInterval : 0
#endif
						 );

		if (pUsbOsPipe->pUsbOsHal->pUsbDevice->speed == USB_SPEED_HIGH) {
                	ule->urb->interval = (pUsbOsPipe->bInterval + 3) * 8;
		}

		ule->urb->transfer_flags |= URB_ASYNC_UNLINK;
	} else {
		usb_fill_bulk_urb(ule->urb, pUsbOsPipe->pUsbOsHal->pUsbDevice, pUsbOsPipe->pipe_os,
						  TransferBuffer,
						  TransferBufferLength,
						  UsbDataCompletionRoutine,
						  ule);
	}
	
	err = SubmitUrbListEntry(ule);
	if(err)
		ChangeUrbListEntryState(ule, ULE_FAILED);
	
	return err;
}

static void UsbControlCompletionRoutine(PURB urb
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
											 , struct pt_regs *regs
#endif
											 )
{
	PURB_LIST_ENTRY ule = urb->context;
	int UsbStatus = urb->status;
	PUINT8 pDataReceived = urb->transfer_buffer;
	int length = urb->actual_length;
	URB_LIST_ENTRY_STATE newstate;
	USB_CTRLREQUEST dr;

	OsDebugPrintf("%s: userCtx=%p ule=%p urb=%p UsbStatus=%d length=%d\n", __FUNCTION__, ule->userContext, ule, urb, UsbStatus, length);

	if(urb->setup_packet)
		memcpy(&dr, urb->setup_packet, sizeof(dr));
	else
		memset(&dr, '\0', sizeof(dr));

	if(ErrorIndicatesDisconnect(UsbStatus)) {
		err("%s: error: URB status %d, surprise disconnect?", __FUNCTION__, UsbStatus);
		ule->pUsbOsHal->disconnected = TRUE;
		ChangeUrbListEntryState(ule, ULE_QUEUED);
		list_add_tail(&ule->queued_entry, &ule->ul->queued_list);
		return;
	}

	switch(UsbStatus) {
		case 0:
			if (urb->pipe & USB_DIR_IN) {
				if(ule->userBufferPtr) {
					memcpy(ule->userBufferPtr, pDataReceived, length);
					pDataReceived = ule->userBufferPtr;
				}
			}
			newstate = ULE_COMPLETED;
			break;
		case -ECONNRESET:
			dbg("%s: -ECONNRESET", __FUNCTION__);
			newstate = ULE_ABORTED;
			break;
		case -ENOENT:
			dbg("%s: -ENOENT", __FUNCTION__);
			newstate = ULE_ABORTED;
			break;
		case -ESHUTDOWN:
			newstate = ULE_FAILED;
			break;
		default:
			err("%s: error: URB status %d", __FUNCTION__, UsbStatus);
			newstate = ULE_FAILED;
			break;
	}

	CallbackOsUsbControlTransferCompleted(ule->userContext, UsbStatus, dr.bRequest, le16_to_cpup(&dr.wValue), le16_to_cpup(&dr.wIndex), pDataReceived, length);
	
	ChangeUrbListEntryState(ule, newstate);
}

__shimcall__
INT OsUsbMakeControlTransfer(PUSBOSHAL pUsbOsHal, BOOLEAN directionIn, OSUSB_TYPE usbType, OSUSB_RECIP usbRecip, PVOID userContext, UINT8 Request, UINT16 Value, UINT16 Index, PVOID TransferBuffer, ULONG TransferBufferLength)
{
	PURB_LIST_ENTRY ule;
	USB_CTRLREQUEST *dr;
	PUSBOSPIPE pUsbOsPipe = directionIn ? pUsbOsHal->pControlInPipe : pUsbOsHal->pControlOutPipe;
	UINT16 wTransferBufferLength = (UINT16)TransferBufferLength;
	int err;

	OsDebugPrintf("%s: pUsbOsPipe=%p/0x%x dir=%s usbType=%d usbRecip=%d\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, (pUsbOsPipe->pipe_os & USB_DIR_IN) ? "IN" : "OUT", usbType, usbRecip);

	ule = GetFreeUrbFromList (&pUsbOsPipe->UrbList);
	if(!ule) {
		OsErrorPrintf("%s: pUsbOsPipe=%p/0x%x no free URBs\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os);

		return -ENOMEM;
	}

	ule->userContext = userContext;

	if (pUsbOsPipe->pipe_os & USB_DIR_IN) {
		if(!IsKMallocMem(TransferBuffer)) {
			ASSERT(nBytes <= ule->bufferSize);
			
			ule->userBufferPtr = TransferBuffer;
			TransferBuffer = ule->bufferPtr;
		} else
			ule->userBufferPtr = NULL;
	} else {
		if(!IsKMallocMem(TransferBuffer)) {
			ASSERT(nBytes <= ule->bufferSize);
			memcpy(ule->bufferPtr, TransferBuffer, TransferBufferLength);
			TransferBuffer = ule->bufferPtr;
		}
	}

	dr = &ule->devRequest;
	dr->bRequestType = (usbType << 5) | usbRecip | ((pUsbOsPipe->pipe_os & USB_DIR_IN) ? USB_DIR_IN : USB_DIR_OUT);
	dr->bRequest = Request;

	dr->wValue = cpu_to_le16p(&Value);
	dr->wIndex = cpu_to_le16p(&Index);
	dr->wLength = cpu_to_le16p(&wTransferBufferLength);

	usb_fill_control_urb(ule->urb, pUsbOsPipe->pUsbOsHal->pUsbDevice, pUsbOsPipe->pipe_os,
			(unsigned char*)dr,
			TransferBuffer,
			TransferBufferLength,
			UsbControlCompletionRoutine,
			ule);

	err = SubmitUrbListEntry(ule);
	if(err)
		ChangeUrbListEntryState(ule, ULE_FAILED);
	
	return err;
}

__shimcall__
PUSBOSPIPE OsUsbCreatePipe(PUSBOSHAL pUsbOsHal, UCHAR bEndpointAddress, UCHAR bmAttributes, UINT16 wMaxPacketSize, UCHAR bInterval)
{
	PUSBOSPIPE pUsbOsPipe;
	
	OsDebugPrintf("%s: pUsbOsHal=%p addr=0x%x attr=0x%x\n", __FUNCTION__, pUsbOsHal, bEndpointAddress, bmAttributes);
	
	pUsbOsPipe = kmalloc(sizeof(USBOSPIPE), GFP_KERNEL);
	if (!pUsbOsPipe) {
		OsErrorPrintf("%s: cannot allocate USBOSPIPE\n", __FUNCTION__);
		return NULL;
	}
	
	memset(pUsbOsPipe, 0, sizeof(USBOSPIPE));
	
	pUsbOsPipe->pUsbOsHal = pUsbOsHal;
	
	pUsbOsPipe->bEndpointAddress = bEndpointAddress;
	pUsbOsPipe->bmAttributes = bmAttributes;
	pUsbOsPipe->wMaxPacketSize = wMaxPacketSize;
	pUsbOsPipe->bInterval = bInterval;
	
	switch(bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) {
		case USB_ENDPOINT_XFER_CONTROL:
			if(bEndpointAddress & USB_DIR_IN)
				pUsbOsPipe->pipe_os = usb_rcvctrlpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
			else
				pUsbOsPipe->pipe_os = usb_sndctrlpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);;
			OsTestPrintf("%s: %p/0x%x CTRL EP%d(%c) addr=0x%x attr=0x%x maxps=%u\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, usb_pipeendpoint(pUsbOsPipe->pipe_os), usb_pipeout(pUsbOsPipe->pipe_os) ? 'O' : 'I', bEndpointAddress, bmAttributes, wMaxPacketSize);
			break;
		case USB_ENDPOINT_XFER_ISOC:
			if(bEndpointAddress & USB_DIR_IN)
				pUsbOsPipe->pipe_os = usb_rcvisocpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
			else
				pUsbOsPipe->pipe_os = usb_sndisocpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);;
			OsTestPrintf("%s: %p/0x%x ISOC EP%d(%c) addr=0x%x attr=0x%x maxps=%u\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, usb_pipeendpoint(pUsbOsPipe->pipe_os), usb_pipeout(pUsbOsPipe->pipe_os) ? 'O' : 'I', bEndpointAddress, bmAttributes, wMaxPacketSize);
			break;
		case USB_ENDPOINT_XFER_BULK:
			if(bEndpointAddress & USB_DIR_IN)
				pUsbOsPipe->pipe_os = usb_rcvbulkpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
			else
				pUsbOsPipe->pipe_os = usb_sndbulkpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);;
			OsTestPrintf("%s: %p/0x%x BULK EP%d(%c) addr=0x%x attr=0x%x maxps=%u\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, usb_pipeendpoint(pUsbOsPipe->pipe_os), usb_pipeout(pUsbOsPipe->pipe_os) ? 'O' : 'I', bEndpointAddress, bmAttributes, wMaxPacketSize);
			break;
		case USB_ENDPOINT_XFER_INT:
			if(bEndpointAddress & USB_DIR_IN)
				pUsbOsPipe->pipe_os = usb_rcvintpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
			else
				pUsbOsPipe->pipe_os = usb_sndintpipe(pUsbOsHal->pUsbDevice, bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);;
			OsTestPrintf("%s: %p/0x%x INTR EP%d(%c) addr=0x%x attr=0x%x maxps=%u\n", __FUNCTION__, pUsbOsPipe, pUsbOsPipe->pipe_os, usb_pipeendpoint(pUsbOsPipe->pipe_os), usb_pipeout(pUsbOsPipe->pipe_os) ? 'O' : 'I', bEndpointAddress, bmAttributes, wMaxPacketSize);
			break;
		default:
			OsDebugPrintf("%s: unsupported type=%d\n", __FUNCTION__, bmAttributes & USB_ENDPOINT_XFERTYPE_MASK);
			kfree(pUsbOsPipe);
			return NULL;
	}

	if(!InitUrbList(&pUsbOsPipe->UrbList, pUsbOsHal, MAX_URB_BUFSIZE, NUM_URBS_PER_PIPE)) {
		FreeUrbList(&pUsbOsPipe->UrbList);
		kfree(pUsbOsPipe);
		return NULL;
	}
	
	list_add_tail(&pUsbOsPipe->entry, &pUsbOsHal->pipe_list);
	
	return pUsbOsPipe;
}

__shimcall__
void OsUsbDestroyPipe (PUSBOSPIPE pUsbOsPipe)
{
	OsDebugPrintf("%s: %p\n", __FUNCTION__, pUsbOsPipe);

	list_del(&pUsbOsPipe->entry);
	AbortListUrbs(&pUsbOsPipe->UrbList);
	FreeUrbList(&pUsbOsPipe->UrbList);

	kfree(pUsbOsPipe);
}

__shimcall__
void OsUsbInit(void)
{
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
 	usbintworkaround = OsConfigOption("workarounds/usbint");
#endif
}

__shimcall__
PUSBOSHAL OsUsbCreateInstance (POS_DEVNODE pDevNode, struct usb_device *pUsbDevice)
{
	PUSBOSHAL			pUsbOsHal;
	
	OsDebugPrintf("%s: pDevNode=%p udev=%p\n", __FUNCTION__, pDevNode, pUsbDevice);
	
	pUsbOsHal = kmalloc(sizeof(USBOSHAL), GFP_KERNEL);
	if (!pUsbOsHal) {
		OsErrorPrintf("%s: cannot allocate USBOSHAL\n", __FUNCTION__);
		return NULL;
	}

	memset(pUsbOsHal, 0, sizeof(USBOSHAL));
	
	pUsbOsHal->pDevNode = pDevNode;
	pUsbOsHal->pUsbDevice = pUsbDevice;

	INIT_LIST_HEAD(&pUsbOsHal->pipe_list);

	pUsbOsHal->pControlOutPipe = OsUsbCreatePipe(pUsbOsHal,
			0,
			USB_ENDPOINT_XFER_CONTROL,
			64,
			0);
	if(!pUsbOsHal->pControlOutPipe) {
		kfree(pUsbOsHal);
		return NULL;
	}

	pUsbOsHal->pControlInPipe = OsUsbCreatePipe(pUsbOsHal,
			0 | USB_DIR_IN,
			USB_ENDPOINT_XFER_CONTROL,
			64,
			0);
	if(!pUsbOsHal->pControlInPipe) {
		OsUsbDestroyPipe(pUsbOsHal->pControlOutPipe);
		kfree(pUsbOsHal);
		return NULL;
	}

	return pUsbOsHal;
}

extern struct semaphore dldr_init_mutex;

__shimcall__
void OsUsbHaltInstance(PUSBOSHAL pUsbOsHal)
{
	struct net_device *ndev = pUsbOsHal->pDevNode->netDev;

	down(&dldr_init_mutex);
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
	if(ndev->reg_state != NETREG_REGISTERED) {
		up(&dldr_init_mutex);
		return;
	}
#endif

	pUsbOsHal->pDevNode->netDev = NULL;
	if(ndev) {
		unregister_netdev(ndev);
	}

	NdisWrapHaltAdapter(pUsbOsHal->pDevNode->miniportAdapter);

	if(ndev) {
		free_netdev(ndev);
	}

	up(&dldr_init_mutex);
}

__shimcall__
void OsUsbDestroyInstance (PUSBOSHAL pUsbOsHal)
{
	struct list_head *e, *n;
	PUSBOSPIPE pUsbOsPipe;
	int i;
	
	OsDebugPrintf("%s: %p\n", __FUNCTION__, pUsbOsHal);

	// destroy all pipes
	list_for_each_safe (e, n, &pUsbOsHal->pipe_list) {
		pUsbOsPipe = list_entry(e, USBOSPIPE, entry);
		OsUsbDestroyPipe(pUsbOsPipe);
	}
	
	for(i = 0; i <
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
			pUsbOsHal->pUsbDevice->actconfig->desc.bNumInterfaces
#else
			pUsbOsHal->pUsbDevice->actconfig->bNumInterfaces
#endif
			; i++) {
		struct usb_interface *intf = usb_ifnum_to_if(pUsbOsHal->pUsbDevice, i);

		if(
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
				usb_get_intfdata(intf)
#else
				intf->private_data
#endif
				== pUsbOsHal->pDevNode) {
			OsTestPrintf("%s: releasing additional interface %d\n", __FUNCTION__, i);
			usb_driver_release_interface(
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
					to_usb_driver(pUsbOsHal->pUsbDevice->dev.driver),
#else
					&dldrusb_driver,
#endif
					intf);
		}
	}
	kfree(pUsbOsHal);
}

__shimcall__
void OsUsbAbortUrbWithContext (PUSBOSHAL pUsbOsHal, PVOID userContext)
{
	struct list_head *h, *e;
	PUSBOSPIPE pUsbOsPipe;
	PURB_LIST_ENTRY ule;
	
	OsDebugPrintf("%s: userContext=%p Irql=%s\n", __FUNCTION__, userContext, OsIrqlName(OsGetIrql()));

	// find Urb
	list_for_each (h, &pUsbOsHal->pipe_list) {
		pUsbOsPipe = list_entry(h, USBOSPIPE, entry);

		list_for_each (e, &pUsbOsPipe->UrbList.urb_list) {
			ule = list_entry(e, URB_LIST_ENTRY, entry);

			//dbg("%s: ule=%p", __FUNCTION__, ule);

			if((ule->userContext == userContext) && (ule->state >= ULE_QUEUED) && (ule->state <= ULE_COMPLETED)) {
				OsDebugPrintf("%s: found ule %p (urb %p) in state %s\n", __FUNCTION__,
						ule, ule->urb, uleStateNames[ule->state]);
				if(ule->state == ULE_SUBMITTED) {
					OsDebugPrintf("%s: unlinking ule %p (urb %p)\n", __FUNCTION__, ule, ule->urb);
					//ule->urb->transfer_flags &= ~URB_ASYNC_UNLINK;
					ule->urb->transfer_flags |= URB_ASYNC_UNLINK;
					usb_unlink_urb(ule->urb);
				} else {
					unsigned long irqlstate;
					OSIRQL oldIrql = OsRaiseIrql(OSIRQL_DISPATCH, &irqlstate);
					if(ule->state == ULE_QUEUED) {
						if((pUsbOsPipe->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_CONTROL) {
							USB_CTRLREQUEST dr;
							if(ule->urb && ule->urb->setup_packet)
								memcpy(&dr, ule->urb->setup_packet, sizeof(dr));
							else
								memset(&dr, '\0', sizeof(dr));

							CallbackOsUsbControlTransferCompleted(ule->userContext, -ENOENT, dr.bRequest, le16_to_cpup(&dr.wValue), le16_to_cpup(&dr.wIndex), NULL, 0);
						} else {
							CallbackOsUsbBulkOrIntTransferCompleted(ule->userContext, -ENOENT, NULL, 0);
						}
					}
					ChangeUrbListEntryState(ule, ULE_ABORTED);
					OsRestoreIrql(oldIrql, &irqlstate);
				}
			}
		}
	}
}

#else

__shimcall__
void OsUsbInit(void)
{
}

__shimcall__ PUSBOSHAL OsUsbCreateInstance (POS_DEVNODE pDevNode, struct usb_device *udev)
{
	return NULL;
}

__shimcall__ void OsUsbHaltInstance(PUSBOSHAL pUsbOsHal)
{
}

__shimcall__ void OsUsbDestroyInstance (PUSBOSHAL pUsbOsHal)
{
}

__shimcall__ PUSBOSPIPE OsUsbCreatePipe(PUSBOSHAL pUsbOsHal, UCHAR bEndpointAddress, UCHAR bmAttributes, UINT16 wMaxPacketSize, UCHAR bInterval)
{
	return NULL;
}

__shimcall__ void OsUsbDestroyPipe (PUSBOSPIPE pUsbOsPipe)
{
}


__shimcall__ INT OsUsbGetDescriptor(PUSBOSHAL pUsbOsHal, UCHAR DescType, UCHAR DescIndex, PVOID TransferBuffer, ULONG TransferBufferLength)
{
	return -ENOSYS;
}

__shimcall__ VOID OsUsbClaimInterface(PUSBOSHAL pUsbOsHal, UCHAR InterfaceNumber, UCHAR AlternateSetting)
{
}

__shimcall__ INT OsUsbSetConfiguration(PUSBOSHAL pUsbOsHal, INT Configuration)
{
	return -ENOSYS;
}

__shimcall__ INT OsUsbResetPort(PUSBOSHAL pUsbOsHal)
{
	return -ENOSYS;
}

__shimcall__ INT OsUsbClearHalt(PUSBOSPIPE pUsbOsPipe)
{
	return -ENOSYS;
}

__shimcall__ INT OsUsbMakeBulkOrIntTransfer(PUSBOSPIPE pUsbOsPipe, PVOID userContext, PVOID TransferBuffer, ULONG TransferBufferLength)
{
	return -ENOSYS;
}

__shimcall__ INT OsUsbMakeControlTransfer(PUSBOSHAL pUsbOsHal, BOOLEAN directionIn, OSUSB_TYPE usbType, OSUSB_RECIP usbRecip, PVOID userContext, UINT8 Request, UINT16 Value, UINT16 Index, PVOID TransferBuffer, ULONG TransferBufferLength)
{
	return -ENOSYS;
}

__shimcall__ void OsUsbAbortUrbWithContext (PUSBOSHAL pUsbOsHal, PVOID userContext)
{
}
#endif // defined(CONFIG_USB) || defined(CONFIG_USB_MODULE)
